##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################

"""
Power Plant costing library
This method leverages NETL costing capabilities. Two main methods have been
developed to calculate the capital cost of power generation plants:

1.- Fossil fueled power plants (from SCPC to IGCC) (get_PP_costing)
2.- supercritical CO2 power cycles (direct and indirect) (get_sCO2_unit_cost)

other methods:
* get_ASU_cost() to cost air separation units
* costing_initialization() to initialize costing blocks
* display_total_plant_costs() to display total plant cost
* display_bare_erected_costs() to display BEC costs
* build_flowsheet_cost_constraint() to display the total cost of the entire
flowsheet
* display_flowsheet_cost() to display flowsheet cost
* check_sCO2_costing_bounds() to display a warnning if costing model have been
used outside the range that where designed for
"""
__author__ = "Costing Team (A. Noring and M. Zamarripa)"
__version__ = "1.0.0"

from pyomo.environ import Param, Var, Block, Constraint, Expression, value, \
    Expr_if
import idaes.core.util.scaling as iscale
from idaes.power_generation.costing.costing_dictionaries import \
    BB_costing_exponents, BB_costing_params, sCO2_costing_params
from pyomo.util.calc_var_value import calculate_variable_from_constraint

# -----------------------------------------------------------------------------
# Power Plant Costing Library
# -----------------------------------------------------------------------------


def get_PP_costing(self, cost_accounts,
                   scaled_param, units, tech, ccs='B'):
    '''
    Power Plant Costing Method
    This method relies on the capital cost scaling methodologies developed
    by NETL. Report #DOE/NETL-341/013113
    Multiple vendors quotes have been used to determine the cost of several
    plant equipments (i.e. boiler, pumps, heat exchangers, etc.), other cost
    incurred during the plant operation (i.e. solids handling, etc.)

    Scaling approach uses one main equation:
        SC = RC*(SP/RP)^Exp
    where:
        SC is the scaled cost
        RC is the reference cost
        SP is the scaled operational parameter
        RP is the reference operational parameter
        Exp is the scaling exponent

    The scaled cost is computed using ref values for different technoligies.
    Categories:
    1 - Supercritical PC, air-fired, with and without CO2 capture,
    Illinois No. 6 coal
    2 - Subcritical PC, air-fired, with and without CO2 capture,
    Illinois No. 6 coal
    3 - Two-stage, slurry-feed, oxygen-blown gasifier with and without
    CO2 capture, Illinois No. 6 coal
    4 - Single-stage, slurry-feed, oxygen-blown gasifier with and without
    CO2 capture, Illinois No. 6 coal
    5 - Single-stage, dry-feed, oxygen-blown, up-flow gasifier with
    and without CO2 capture, Illinois No. 6 coal
    6 - Natural gas, air-fired, with and without CO2 capture
    7 - Advanced Ultrasupercritical PC

    This method computes the capital cost of units and main components of the
    power plant, and requires a few arguments to build a constraint as part of
    your main model.

    Args:
    * self: A block or unit model where costing constraints can be added to
    * accounts: A list of accounts to be included in the total cost,
    they should all use the same reference parameter
    * scaled_param: the process parameter for the system(s) being costed
    * units: the units of the scaled_param, used for verification
    * tech: int 1-7 representing the above catagories
    * ccs: 'A' or 'B' representing no CCS or CCS

    The appropriate scaling parameters for various cost accounts can be found
    in the QGESS on capital cost scaling (Report #DOE/NETL-341/013113).
    The correct units for the reference parameters are found in the BBR4 COE
    spreadsheet.

    '''
    # ------------------------ Power Plant Cost ------------------------

    # check to see if a costing block already exists
    if hasattr(self, 'costing'):
        raise AttributeError("{} already has an attribute costing. "
                             "Check that you are not calling get_costing"
                             " twice on the same model".format(self.name))

    # create a costing Block
    self.costing = Block()
    self.costing.library = 'PP'

    # find flowsheet block to create global costing parameters
    try:
        fs = self.flowsheet()
    except AttributeError:
        fs = self.parent_block()

    # build flowsheet level parameters CE_index = year 
    if not hasattr(fs, 'costing'):
        fs.get_costing(year='2018')

    CE_index = fs.costing.CE_index

    # define preloaded accounts
    PC_preloaded_accounts = {'Coal Handling': ['1.1', '1.2',
                                               '1.3', '1.4', '1.9a'],
                             'Sorbent Handling': ['1.5', '1.6',
                                                  '1.7', '1.8', '1.9b'],
                             'Coal Feed': ['2.1', '2.2', '2.9a'],
                             'Sorbent Feed': ['2.5', '2.6', '2.9b'],
                             'Feedwater System': ['3.1', '3.3'],
                             'PC Boiler': ['4.9'],
                             'Steam Turbine': ['8.1'],
                             'Condenser': ['8.3'],
                             'Cooling Tower': ['9.1'],
                             'Circulating Water System': ['9.2', '9.3',
                                                          '9.4', '9.6', '9.7'],
                             'Ash Handling': ['10.6', '10.7', '10.9']}

    IGCC_preloaded_accounts = {'Coal Handling': ['1.1', '1.2',
                                                 '1.3', '1.4', '1.9'],
                               'Coal Feed': ['2.1', '2.2',
                                             '2.3', '2.4', '2.9'],
                               'Feedwater System': ['3.1', '3.3'],
                               'Gasifier': ['4.1'],
                               'Syngas Cooler': ['4.2'],
                               'ASU': ['4.3a'],
                               'ASU Oxidant Compression': ['4.3b'],
                               'Combustion Turbine': ['6.1', '6.3'],
                               'Syngas Expander': ['6.2'],
                               'HRSG': ['7.1', '7.2'],
                               'Steam Turbine': ['8.1'],
                               'Condenser': ['8.3'],
                               'Cooling Tower': ['9.1'],
                               'Circulating Water System': ['9.2', '9.3',
                                                            '9.4', '9.6',
                                                            '9.7'],
                               'Slag Handling': ['10.1', '10.2',
                                                 '10.3', '10.6',
                                                 '10.7', '10.8',
                                                 '10.9']}

    NGCC_preloaded_accounts = {'Feedwater System': ['3.1', '3.3'],
                               'Combustion Turbine': ['6.1', '6.3'],
                               'HRSG': ['7.1', '7.2'],
                               'Steam Turbine': ['8.1'],
                               'Condenser': ['8.3'],
                               'Cooling Tower': ['9.1'],
                               'Circulating Water System': ['9.2', '9.3',
                                                            '9.4', '9.6',
                                                            '9.7']}

    AUSC_preloaded_accounts = {'PC Boiler': ['4.9'],
                               'Steam Turbine': ['8.1'],
                               'Steam Piping': ['8.4']}

    # preloaded account handling
    if type(cost_accounts) == str:
        if tech in [1, 2]:
            cost_accounts = PC_preloaded_accounts[cost_accounts]
        elif tech in [3, 4, 5]:
            cost_accounts = IGCC_preloaded_accounts[cost_accounts]
        elif tech == 6:
            cost_accounts = NGCC_preloaded_accounts[cost_accounts]
        elif tech == 7:
            cost_accounts = AUSC_preloaded_accounts[cost_accounts]
        else:
            AttributeError("{} technology not supported".format(self.name))

    # check that all accounts use the same process parameter
    param_check = None
    for account in cost_accounts:
        param = BB_costing_exponents[str(tech)][account]['Process Parameter']
        if param_check is None:
            param_check = param
        elif param != param_check:
            raise ValueError("{} cost accounts selected do not use "
                             " the same process parameter".format(self.name))

    # check that the user passed the correct units
    ref_units = BB_costing_params[str(tech)][ccs][cost_accounts[0]]['Units']
    if units != ref_units:
        raise ValueError('Account %s uses units of %s. '
                         'Units of %s were passed.'
                         % (cost_accounts[0], ref_units, units))

    # construct dictionaries
    account_names = {}
    exponents = {}
    reference_costs = {}
    reference_params = {}
    engineering_fees = {}
    process_contingencies = {}
    project_contingencies = {}

    for account in cost_accounts:
        account_names[account] = BB_costing_exponents[str(
            tech)][account]['Account Name']
        exponents[account] = float(
            BB_costing_exponents[str(tech)][account]['Exponent'])
        reference_costs[account] = BB_costing_params[str(
            tech)][ccs][account]['BEC']
        reference_params[account] = BB_costing_params[str(
            tech)][ccs][account]['RP Value']
        engineering_fees[account] = BB_costing_params[str(
            tech)][ccs][account]['Eng Fee']
        process_contingencies[account] = BB_costing_params[str(
            tech)][ccs][account]['Process Contingency']
        project_contingencies[account] = BB_costing_params[str(
            tech)][ccs][account]['Project Contingency']

    # Used by other functions for reporting results
    self.costing.account_names = account_names

    # define parameters
    self.costing.exp = Param(cost_accounts,
                             mutable=True,
                             initialize=exponents,
                             doc='exponential parameter for account')

    self.costing.ref_cost = Param(cost_accounts,
                                  mutable=True,
                                  initialize=reference_costs,
                                  doc='reference cost for account')

    self.costing.ref_param = Param(cost_accounts,
                                   mutable=True,
                                   initialize=reference_params,
                                   doc='reference parameter for account')

    self.costing.eng_fee = Param(cost_accounts,
                                 mutable=True,
                                 initialize=engineering_fees,
                                 doc='engineering fee percentage')

    self.costing.process_conting = Param(cost_accounts,
                                         mutable=True,
                                         initialize=process_contingencies,
                                         doc='process contingency percentage')

    self.costing.project_conting = Param(cost_accounts,
                                         mutable=True,
                                         initialize=project_contingencies,
                                         doc='project contingency percentage')

    # define variables
    self.costing.bare_erected_cost = Var(cost_accounts,
                                         initialize=reference_costs,
                                         bounds=(0, 1e4),
                                         doc='scaled bare erected cost in $MM')

    self.costing.total_plant_cost = Var(cost_accounts,
                                        initialize=reference_costs,
                                        bounds=(0, 1e4),
                                        doc='total plant cost in $MM')

    # rule for scaling BEC
    # reference cost is in 2018 dollars, 671.1 is CE index for 2018
    def bare_erected_cost_rule(costing, i):
        return (costing.bare_erected_cost[i]*1e3 ==
                (CE_index/671.1)*costing.ref_cost[i] *
                (scaled_param/costing.ref_param[i])**costing.exp[i])

    self.costing.bare_erected_cost_eq = Constraint(
        cost_accounts, rule=bare_erected_cost_rule)

    # rule for calculating TPC
    def total_plant_cost_rule(costing, i):
        return (costing.total_plant_cost[i] == costing.bare_erected_cost[i] *
                (1 + costing.eng_fee[i] + costing.process_conting[i]) *
                (1 + costing.project_conting[i]))

    self.costing.total_plant_cost_eq = Constraint(
        cost_accounts, rule=total_plant_cost_rule)

    # rule for sum of BEC
    def BEC_sum_rule(costing):
        return sum(costing.bare_erected_cost[i] for i in cost_accounts)
    self.costing.bare_erected_cost_sum = Expression(rule=BEC_sum_rule)

    # rule for sum of TPC
    def TPC_sum_rule(costing):
        return sum(costing.total_plant_cost[i] for i in cost_accounts)
    self.costing.total_plant_cost_sum = Expression(rule=TPC_sum_rule)

    #  # add variable and constraint scaling
    for i in cost_accounts:
        iscale.set_scaling_factor(self.costing.bare_erected_cost[i], 1)
        iscale.set_scaling_factor(self.costing.total_plant_cost[i], 1)
        iscale.constraint_scaling_transform(self.
                                            costing.bare_erected_cost_eq[i],
                                            1e-3,
                                            overwrite=False)
        iscale.constraint_scaling_transform(self.
                                            costing.total_plant_cost_eq[i],
                                            1,
                                            overwrite=False)


# -----------------------------------------------------------------------------
# Supercritical CO2 Costing Library
# -----------------------------------------------------------------------------
def get_sCO2_unit_cost(self, equipment, scaled_param, temp_C=None, n_equip=1):
    '''
    Args:
    self - pyomo Block where constraints will be made
    unit_name - the name of the SCO2 equipment to cost
    scaling_param - the scaling parameter (in appropriate units) for the
                    selected equipment
    temp - the maximum temperature of the equipment. Not all types of
           equipment use a temperature correction factor, so it is optional
    n_equip - the number of pieces of equipment to cost

    Cost is in M$
    '''
    # check to see if a costing block already exists
    if hasattr(self, 'costing'):
        raise AttributeError("{} already has an attribute costing. "
                             "Check that you are not calling get_costing"
                             " twice on the same model".format(self.name))

    # create a costing Block
    self.costing = Block()
    self.costing.library = 'sCO2'
    self.costing.equipment = equipment

    # find flowsheet block to create global costing parameters
    try:
        fs = self.flowsheet()
    except AttributeError:
        fs = self.parent_block()

    # build flowsheet level parameters CE_index = year 
    if not hasattr(fs, 'costing'):
        fs.get_costing(year='2017')

    CE_index = fs.costing.CE_index

    param_dict = sCO2_costing_params[equipment]

    # define parameters
    self.costing.ref_cost = Param(mutable=True,
                                  initialize=param_dict['a'],
                                  doc='Reference cost')

    self.costing.exp = Param(mutable=True,
                             initialize=param_dict['b'],
                             doc='Scaling exponent')

    self.costing.c = Param(mutable=True,
                           initialize=param_dict['c'],
                           doc='coefficient for temperature correction')

    self.costing.d = Param(mutable=True,
                           initialize=param_dict['d'],
                           doc='coefficient for temperature correction')

    self.costing.material_cost = Param(mutable=True,
                                       doc='material installation cost',
                                       initialize=param_dict['Material Cost'])

    self.costing.labor_cost = Param(mutable=True,
                                    initialize=param_dict['Labor Cost'],
                                    doc='labor installation cost')

    # estimates for the percentages of TPC will be added later

    self.costing.eng_fee = Param(mutable=True,
                                 initialize=0,
                                 doc='engineering fee percentage')

    self.costing.process_conting = Param(mutable=True,
                                         initialize=0,
                                         doc='process contingency percentage')

    self.costing.project_conting = Param(mutable=True,
                                         initialize=0,
                                         doc='project contingency percentage')

    # define variables
    # n_equip is left as a fixed variable to support MINLP optimization
    self.costing.n_equip = Var(initialize=n_equip,
                               doc='number of pieces of equipment')

    self.costing.n_equip.fix(n_equip)

    self.costing.scaled_param = Var(initialize=scaled_param,
                                    bounds=(0, 1e12),
                                    doc='scaled parameter')

    self.costing.temp_factor = Var(initialize=1,
                                   bounds=(0.9, 100),
                                   doc='temperature correction factor')

    self.costing.equipment_cost = Var(initialize=self.costing.ref_cost,
                                      bounds=(0, 1e4),
                                      doc='equipment cost of sCO2 unit in $MM')

    self.costing.bare_erected_cost = Var(initialize=self.costing.ref_cost,
                                         bounds=(0, 1e4),
                                         doc='bare erected cost of sCO2 unit'
                                         'in $MM')

    self.costing.total_plant_cost = Var(initialize=self.costing.ref_cost,
                                        bounds=(0, 1e4),
                                        doc='total plant cost of sCO2 unit'
                                        'in $MM')

    # divides the scaled parameter by the number of pieces of equipment
    def scaled_param_rule(costing):
        return costing.scaled_param*costing.n_equip == scaled_param

    self.costing.scaled_param_eq = Constraint(rule=scaled_param_rule)

    # check if equipment requires a temperature correction factor
    if equipment in ['Axial turbine', 'Radial turbine', 'Coal-fired heater',
                     'Natural gas-fired heater', 'Recuperator']:

        if temp_C is None:
            raise ValueError('Temperature argument is '
                             'required to cost %s equipment' % equipment)

        else:
            self.costing.temperature = Var(initialize=500,
                                           bounds=(0, 1e6),
                                           doc='dummy var for temperature')

            self.costing.temp_eq = Constraint(expr=(self.costing.temperature
                                                    == temp_C))

            def temp_correction_rule(costing):  # rule for temp correction
                return (Expr_if(costing.temperature < 550,
                                1e-6*costing.temperature + 1,
                                1 + costing.c*(costing.temperature - 550)
                                + costing.d*(costing.temperature - 550)**2) ==
                        costing.temp_factor)

            self.costing.temp_correction_eq = Constraint(
                rule=temp_correction_rule)
    else:
        self.costing.temp_factor.fix(1)

    # rule for equipment cost
    def equipment_cost_rule(costing):
        return (costing.equipment_cost*1e6 ==
                (CE_index/567.5) * costing.n_equip * costing.ref_cost *
                (costing.scaled_param**costing.exp) * costing.temp_factor)

    self.costing.equipment_cost_eq = Constraint(rule=equipment_cost_rule)

    # rule for bare erected cost
    def bare_erected_cost_rule(costing):
        return (costing.bare_erected_cost == costing.equipment_cost *
                (1 + costing.material_cost + costing.labor_cost))

    self.costing.bare_erected_cost_eq = Constraint(rule=bare_erected_cost_rule)

    # rule for calculating total plant cost
    def total_plant_cost_rule(costing):
        return (costing.total_plant_cost == costing.bare_erected_cost *
                (1 + costing.eng_fee + costing.process_conting
                 + costing.project_conting))
    self.costing.total_plant_cost_eq = Constraint(rule=total_plant_cost_rule)

    # add variable and constraint scaling
    if equipment in ["Recuperator", "Direct air cooler"]:
        iscale.set_scaling_factor(self.costing.scaled_param, 1e-5)
    else:
        iscale.set_scaling_factor(self.costing.scaled_param, 1)

    iscale.set_scaling_factor(self.costing.equipment_cost, 1e3)
    iscale.set_scaling_factor(self.costing.bare_erected_cost, 1e3)
    iscale.set_scaling_factor(self.costing.total_plant_cost, 1e3)
    iscale.constraint_scaling_transform(
        self.costing.equipment_cost_eq, 1e-6, overwrite=False)
    iscale.constraint_scaling_transform(
        self.costing.bare_erected_cost_eq, 1e3, overwrite=False)
    iscale.constraint_scaling_transform(
        self.costing.bare_erected_cost_eq, 1e3, overwrite=False)


# -----------------------------------------------------------------------------
# Air Separation Unit Costing Library
# -----------------------------------------------------------------------------
def get_ASU_cost(self, scaled_param):
    # scaled parameter is O2 flowrate in TPD

    params = {'Reference Cost': 3.26e6,
              'Reference Parameter': 13078,
              'Exponent': 0.7,
              'Eng Fee': 0.097,
              'Process': 0,
              'Project': 0.110}

    # check to see if a costing block already exists
    if hasattr(self, 'costing'):
        raise AttributeError("{} already has an attribute costing. "
                             "Check that you are not calling get_costing"
                             " twice on the same model".format(self.name))

    # create a costing Block
    self.costing = Block()
    self.costing.library = 'ASU'

    # find flowsheet block to create global costing parameters
    try:
        fs = self.flowsheet()
    except AttributeError:
        fs = self.parent_block()

    # build flowsheet level parameters CE_index = year 
    if not hasattr(fs, 'costing'):
        fs.get_costing(year='2017')

    CE_index = fs.costing.CE_index

    # define parameters
    self.costing.ref_cost = Param(initialize=params['Reference Cost'],
                                  mutable=True,
                                  doc='ASU reference cost')

    self.costing.ref_param = Param(initialize=params['Reference Parameter'],
                                   mutable=True,
                                   doc='ASU reference parameter value')

    self.costing.exp = Param(initialize=params['Exponent'],
                             mutable=True,
                             doc='ASU scaling exponent')

    self.costing.eng_fee = Param(mutable=True,
                                 initialize=params['Eng Fee'],
                                 doc='engineering fee percentage')

    self.costing.process_conting = Param(mutable=True,
                                         initialize=params['Process'],
                                         doc='process contingency percentage')

    self.costing.project_conting = Param(mutable=True,
                                         initialize=params['Project'],
                                         doc='project contingency percentage')

    # define variables
    self.costing.bare_erected_cost = Var(initialize=params['Reference Cost'],
                                         bounds=(0, 1e4),
                                         doc='scaled bare erected cost in $MM')

    self.costing.total_plant_cost = Var(initialize=params['Reference Cost'],
                                        bounds=(0, 1e4),
                                        doc='total plant cost in $MM')

    # rule for scaling BEC
    # reference cost is in 2008 dollars, 566.2 is CE index for Nov 2008

    def bare_erected_cost_rule(costing):
        return (costing.bare_erected_cost*1e3 ==
                (CE_index/566.2)*costing.ref_cost *
                (scaled_param/costing.ref_param)**costing.exp)

    self.costing.bare_erected_cost_eq = Constraint(rule=bare_erected_cost_rule)

    # rule for calculating TPC
    def total_plant_cost_rule(costing):
        return (costing.total_plant_cost == costing.bare_erected_cost *
                (1 + costing.eng_fee + costing.process_conting
                 + costing.project_conting))
    self.costing.total_plant_cost_eq = Constraint(rule=total_plant_cost_rule)

    # add variable and constraint scaling
    iscale.set_scaling_factor(self.costing.bare_erected_cost, 1)
    iscale.set_scaling_factor(self.costing.total_plant_cost, 1)

    iscale.constraint_scaling_transform(
        self.costing.bare_erected_cost_eq, 1e-3, overwrite=False)
    iscale.constraint_scaling_transform(
        self.costing.total_plant_cost_eq, 1, overwrite=False)


# -----------------------------------------------------------------------------
# Costing Library Utility Functions
# -----------------------------------------------------------------------------
def costing_initialization(fs):
    for o in fs.component_objects(descend_into=False):
        # look for costing blocks
        if hasattr(o, 'costing'):
            if o.costing.library == 'sCO2':

                if o.costing.equipment in ['Axial turbine',
                                           'Radial turbine',
                                           'Coal-fired heater',
                                           'Natural gas-fired heater',
                                           'Recouperator']:
                    calculate_variable_from_constraint(o.costing.temperature,
                                                       o.costing.temp_eq)
                    calculate_variable_from_constraint(o.costing.temp_factor,
                                                       o.costing.
                                                       temp_correction_eq)

                calculate_variable_from_constraint(o.costing.scaled_param,
                                                   o.costing.scaled_param_eq)
                calculate_variable_from_constraint(o.costing.equipment_cost,
                                                   o.costing.equipment_cost_eq)
                calculate_variable_from_constraint(o.costing.bare_erected_cost,
                                                   o.costing.
                                                   bare_erected_cost_eq)
                calculate_variable_from_constraint(o.costing.total_plant_cost,
                                                   o.costing.
                                                   total_plant_cost_eq)

            elif o.costing.library in ['PP', 'ASU']:
                for key in o.costing.bare_erected_cost.keys():
                    calculate_variable_from_constraint(o.costing.
                                                       bare_erected_cost[key],
                                                       o.costing.
                                                       bare_erected_cost_eq[
                                                           key])
                    calculate_variable_from_constraint(o.costing.
                                                       total_plant_cost[key],
                                                       o.costing.
                                                       total_plant_cost_eq[
                                                           key])


def display_total_plant_costs(fs):
    print('-----Total Plant Costs-----')
    for o in fs.component_objects(descend_into=False):
        # look for costing blocks
        if hasattr(o, 'costing') and hasattr(o.costing, 'total_plant_cost'):
            print('%s: $%.2f Million' % (value(o.name),
                                         value(o.costing.total_cost)))


def display_bare_erected_costs(fs):
    print('-----Bare Erected Costs-----')
    for o in fs.component_objects(descend_into=False):
        # look for costing blocks
        if hasattr(o, 'costing') and hasattr(o.costing, 'bare_erected_cost'):
            print('%s: $%.2f Million' % (value(o.name),
                                         value(o.costing.bare_erected_cost)))


def display_equipment_costs(fs):
    print('-----Equipment Costs-----')
    for o in fs.component_objects(descend_into=False):
        # look for costing blocks
        if hasattr(o, 'costing') and hasattr(o.costing, 'equipment_cost'):
            print('%s: $%.2f Million' % (value(o.name),
                                         value(o.costing.equipment_cost)))


def build_flowsheet_cost_constraint(m):
    total_cost_list = []
    for o in m.fs.component_objects(descend_into=False):
        # look for costing blocks
        if hasattr(o, 'costing'):
            for key in o.costing.total_plant_cost.keys():
                total_cost_list.append(o.costing.total_plant_cost[key])

    m.fs.flowsheet_cost = Var(initialize=0,
                              bounds=(0, 1e12),
                              doc='cost of entire process')

    def flowsheet_cost_rule(fs):
        return fs.flowsheet_cost == sum(total_cost_list)

    m.fs.flowsheet_cost_eq = Constraint(rule=flowsheet_cost_rule)


def display_flowsheet_cost(m):
    print('\n')
    print('Total flowsheet cost: $%.3f Million' %
          value(m.fs.flowsheet_cost_exp))


def check_sCO2_costing_bounds(fs):
    # interate through the children of the flowsheet
    for o in fs.component_objects(descend_into=False):
        # look for costing blocks
        if hasattr(o, 'costing'):
            costing = o.costing
            if costing.library == 'sCO2':
                equipment = costing.equipment
                lower_bound = sCO2_costing_params[equipment]['Lower Bound']
                upper_bound = sCO2_costing_params[equipment]['Upper Bound']
                if value(costing.scaled_param) < lower_bound:
                    print('''%s: The scaled parameter (%f) is below the lower
                          bound (%f).''' % (value(o.name),
                                            value(costing.scaled_param),
                                            lower_bound))
                elif value(costing.scaled_param) > upper_bound:
                    print('''%s: The scaled parameter (%f) is above the upper
                          bound (%f).''' % (value(o.name),
                                            value(costing.scaled_param),
                                            upper_bound))
                else:
                    print('''%s: The scaled parameter is within the bounds.'''
                          % value(o.name))



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Methods for calculating pure component properties from:

The Properties of Gases & Liquids, 4th Edition
Reid, Prausnitz and Polling, 1987, McGraw-Hill

All parameter indicies and units based on conventions used by the source
"""
from pyomo.environ import exp, log, Var, units as pyunits

from idaes.core.util.misc import set_param_from_config


# -----------------------------------------------------------------------------
# Heat capacities, enthalpies and entropies
class cp_mol_ig_comp():

    @staticmethod
    def build_parameters(cobj):
        cobj.cp_mol_ig_comp_coeff_A = Var(
                doc="Parameter A for ideal gas molar heat capacity",
                units=pyunits.J/pyunits.mol/pyunits.K)
        set_param_from_config(cobj, param="cp_mol_ig_comp_coeff", index="A")

        cobj.cp_mol_ig_comp_coeff_B = Var(
                doc="Parameter B for ideal gas molar heat capacity",
                units=pyunits.J/pyunits.mol/pyunits.K**2)
        set_param_from_config(cobj, param="cp_mol_ig_comp_coeff", index="B")

        cobj.cp_mol_ig_comp_coeff_C = Var(
                doc="Parameter C for ideal gas molar heat capacity",
                units=pyunits.J/pyunits.mol/pyunits.K**3)
        set_param_from_config(cobj, param="cp_mol_ig_comp_coeff", index="C")

        cobj.cp_mol_ig_comp_coeff_D = Var(
                doc="Parameter D for ideal gas molar heat capacity",
                units=pyunits.J/pyunits.mol/pyunits.K**4)
        set_param_from_config(cobj, param="cp_mol_ig_comp_coeff", index="D")

    @staticmethod
    def return_expression(b, cobj, T):
        # Specific heat capacity
        T = pyunits.convert(T, to_units=pyunits.K)

        cp = (cobj.cp_mol_ig_comp_coeff_D*T**3 +
              cobj.cp_mol_ig_comp_coeff_C*T**2 +
              cobj.cp_mol_ig_comp_coeff_B*T +
              cobj.cp_mol_ig_comp_coeff_A)

        units = b.params.get_metadata().derived_units
        return pyunits.convert(cp, units["heat_capacity_mole"])


class enth_mol_ig_comp():

    @staticmethod
    def build_parameters(cobj):
        if not hasattr(cobj, "cp_mol_ig_comp_coeff_A"):
            cp_mol_ig_comp.build_parameters(cobj)

        if cobj.parent_block().config.include_enthalpy_of_formation:
            units = cobj.parent_block().get_metadata().derived_units

            cobj.enth_mol_form_vap_comp_ref = Var(
                    doc="Vapor phase molar heat of formation @ Tref",
                    units=units["energy_mole"])
            set_param_from_config(cobj, param="enth_mol_form_vap_comp_ref")

    @staticmethod
    def return_expression(b, cobj, T):
        # Specific enthalpy
        T = pyunits.convert(T, to_units=pyunits.K)
        Tr = pyunits.convert(b.params.temperature_ref, to_units=pyunits.K)

        units = b.params.get_metadata().derived_units

        h_form = (cobj.enth_mol_form_vap_comp_ref if
                  b.params.config.include_enthalpy_of_formation
                  else 0*units["energy_mole"])

        h = (pyunits.convert(
                (cobj.cp_mol_ig_comp_coeff_D/4)*(T**4-Tr**4) +
                (cobj.cp_mol_ig_comp_coeff_C/3)*(T**3-Tr**3) +
                (cobj.cp_mol_ig_comp_coeff_B/2)*(T**2-Tr**2) +
                cobj.cp_mol_ig_comp_coeff_A*(T-Tr), units["energy_mole"]) +
             h_form)

        return h


class entr_mol_ig_comp():

    @staticmethod
    def build_parameters(cobj):
        if not hasattr(cobj, "cp_mol_ig_comp_coeff_A"):
            cp_mol_ig_comp.build_parameters(cobj)

        units = cobj.parent_block().get_metadata().derived_units

        cobj.entr_mol_form_vap_comp_ref = Var(
                doc="Vapor phase molar entropy of formation @ Tref",
                units=units["entropy_mole"])
        set_param_from_config(cobj, param="entr_mol_form_vap_comp_ref")

    @staticmethod
    def return_expression(b, cobj, T):
        # Specific entropy
        T = pyunits.convert(T, to_units=pyunits.K)
        Tr = pyunits.convert(b.params.temperature_ref, to_units=pyunits.K)

        units = b.params.get_metadata().derived_units

        s = (pyunits.convert(
                (cobj.cp_mol_ig_comp_coeff_D/3)*(T**3-Tr**3) +
                (cobj.cp_mol_ig_comp_coeff_C/2)*(T**2-Tr**2) +
                cobj.cp_mol_ig_comp_coeff_B*(T-Tr) +
                cobj.cp_mol_ig_comp_coeff_A*log(T/Tr),
                units["entropy_mole"]) +
             cobj.entr_mol_form_vap_comp_ref)

        return s


# -----------------------------------------------------------------------------
# Saturation pressure
# Note that this equation in not valid beyond the critical temperature
class pressure_sat_comp():

    @staticmethod
    def build_parameters(cobj):
        cobj.pressure_sat_comp_coeff_A = Var(
                doc="Coefficient A for calculating Psat",
                units=pyunits.dimensionless)
        set_param_from_config(cobj, param="pressure_sat_comp_coeff", index="A")

        cobj.pressure_sat_comp_coeff_B = Var(
                doc="Coefficient B for calculating Psat",
                units=pyunits.dimensionless)
        set_param_from_config(cobj, param="pressure_sat_comp_coeff", index="B")

        cobj.pressure_sat_comp_coeff_C = Var(
                doc="Coefficient C for calculating Psat",
                units=pyunits.dimensionless)
        set_param_from_config(cobj, param="pressure_sat_comp_coeff", index="C")

        cobj.pressure_sat_comp_coeff_D = Var(
                doc="Coefficient D for calculating Psat",
                units=pyunits.dimensionless)
        set_param_from_config(cobj, param="pressure_sat_comp_coeff", index="D")

    @staticmethod
    def return_expression(b, cobj, T, dT=False):
        if dT:
            return pressure_sat_comp.dT_expression(b, cobj, T)

        x = 1 - T/cobj.temperature_crit

        return (exp((1-x)**-1 * (cobj.pressure_sat_comp_coeff_A*x +
                                 cobj.pressure_sat_comp_coeff_B*x**1.5 +
                                 cobj.pressure_sat_comp_coeff_C*x**3 +
                                 cobj.pressure_sat_comp_coeff_D*x**6)) *
                cobj.pressure_crit)

    @staticmethod
    def dT_expression(b, cobj, T):
        x = 1 - T/cobj.temperature_crit

        return (-pressure_sat_comp.return_expression(b, cobj, T) *
                ((cobj.pressure_sat_comp_coeff_A +
                  1.5*cobj.pressure_sat_comp_coeff_B*x**0.5 +
                  3*cobj.pressure_sat_comp_coeff_C*x**2 +
                  6*cobj.pressure_sat_comp_coeff_D*x**5)/T +
                 (cobj.temperature_crit/T**2) *
                 (cobj.pressure_sat_comp_coeff_A*x +
                  cobj.pressure_sat_comp_coeff_B*x**1.5 +
                  cobj.pressure_sat_comp_coeff_C*x**3 +
                  cobj.pressure_sat_comp_coeff_D*x**6)))


# -----------------------------------------------------------------------------
class RPP4(object):
    cp_mol_ig_comp = cp_mol_ig_comp
    enth_mol_ig_comp = enth_mol_ig_comp
    entr_mol_ig_comp = entr_mol_ig_comp
    pressure_sat_comp = pressure_sat_comp



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""Test the basic 0 = z^3 + bz^2 + cz + d calculator."""

import os
import pytest
import pyomo.environ as pyo
import idaes

def cubic_function(z, b, c, d):
    return z**3 + b*z**2 + c*z + d

def derivs(z, b, c, d):
    # Use the tested derivative formulas to make sure the external function
    # keeps returning the right answers. These formulas were confirmed and
    # documented in a jupyter notebook.
    g = [None]*3
    h = [None]*6
    g[0] = 1.0/(-1 + c/z**2 + 2*d/z**3)
    g[1] = 1.0/(-2*z - b + d/z**2)
    g[2] = 1.0/(-3*z**2 - 2*b*z - c)
    """ the hessian structure is:
            b c d
          +------
        b | 0 1 3
        c | - 2 4
        d | - - 5
    """
    h[0] = g[0]**3*(2*c/z**3 + 6*d/z**4)
    h[1] = g[0]**2*(2*c/z**3*g[1] + 6*d/z**4*g[1] - 1/z**2)
    h[2] = g[1]**3*(2 + 2*d/z**3)
    h[3] = g[0]**2*(2*c/z**3*g[2] + 6*d/z**4*g[2] - 2/z**3)
    h[4] = g[1]**2*(2*g[2] + 2*d/z**3*g[2] - 1/z**2)
    h[5] = g[2]**3*(6*z + 2*b)
    return g, h


@pytest.mark.unit
def test_general_cubic_root_finder():
    plib = os.path.join(idaes.bin_directory, "cubic_roots.so")
    m = pyo.ConcreteModel()
    m.croot_l = pyo.ExternalFunction(library=plib, function="cubic_root_l")
    m.croot_m = pyo.ExternalFunction(library=plib, function="cubic_root_m")
    m.croot_h = pyo.ExternalFunction(library=plib, function="cubic_root_h")
    param_dict = {
        1:{"b":-3, "c":0.5, "d":-1, "three":False},
        2:{"b":-3, "c":0.5, "d":1, "three":True},
        3:{"b":1, "c":0.5, "d":-1, "three":False},
        4:{"b":1, "c":0.5, "d":2, "three":False},
    }

    for k, v in param_dict.items():
        b = v["b"]
        c = v["c"]
        d = v["d"]
        fl, gl, hl = m.croot_l.evaluate_fgh(args=(b, c, d))
        fm, gm, hm = m.croot_m.evaluate_fgh(args=(b, c, d))
        fh, gh, hh = m.croot_h.evaluate_fgh(args=(b, c, d))
        if v["three"]:
            assert fl < fm
            assert fm < fh
        else:
            assert pytest.approx(fl, abs=1e-5) == fm
            assert pytest.approx(fm, abs=1e-5) == fh
        assert pytest.approx(0, abs=1e-6) == cubic_function(fl, b, c, d)
        assert pytest.approx(0, abs=1e-6) == cubic_function(fm, b, c, d)
        assert pytest.approx(0, abs=1e-6) == cubic_function(fh, b, c, d)
        g, h = derivs(fl, b, c, d)
        for i, ge in enumerate(g):
            assert pytest.approx(ge, abs=1e-5) == gl[i]
        for i, he in enumerate(h):
            assert pytest.approx(he, abs=1e-5) == hl[i]
        g, h = derivs(fm, b, c, d)
        for i, ge in enumerate(g):
            assert pytest.approx(ge, abs=1e-5) == gm[i]
        for i, he in enumerate(h):
            assert pytest.approx(he, abs=1e-5) == hm[i]
        g, h = derivs(fh, b, c, d)
        for i, ge in enumerate(g):
            assert pytest.approx(ge, abs=1e-5) == gh[i]
        for i, he in enumerate(h):
            assert pytest.approx(he, abs=1e-5) == hh[i]



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for tray column unit model (single feed tray, no side draws).

Author: Jaffer Ghouse
"""
import pytest
from pyomo.environ import (ConcreteModel, TerminationCondition,
                           SolverStatus, value)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.generic_models.unit_models.distillation import TrayColumn
from idaes.generic_models.unit_models.distillation.condenser \
    import CondenserType, TemperatureSpec
from idaes.generic_models.properties.activity_coeff_models.\
    BTX_activity_coeff_VLE import BTXParameterBlock
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import \
    PhysicalParameterTestBlock, initialization_tester
from idaes.core.util import get_solver

from idaes.generic_models.properties.core.generic.generic_property \
    import GenericParameterBlock

from idaes.generic_models.properties.core.examples.BT_ideal \
    import configuration

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


@pytest.mark.unit
def test_config():

    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.properties = PhysicalParameterTestBlock()

    m.fs.unit = TrayColumn(default={
                           "number_of_trays": 10,
                           "feed_tray_location": 5,
                           "condenser_type": CondenserType.totalCondenser,
                           "condenser_temperature_spec":
                               TemperatureSpec.customTemperature,
                           "property_package": m.fs.properties,
                           "has_heat_transfer": False,
                           "has_pressure_change": False})

    assert len(m.fs.unit.config) == 12

    assert m.fs.unit.feed_tray.config.is_feed_tray

    assert hasattr(m.fs.unit, "condenser")
    assert hasattr(m.fs.unit, "reboiler")

    assert hasattr(m.fs.unit, "rectification_section")
    assert hasattr(m.fs.unit, "stripping_section")


class TestBTXIdeal():
    @pytest.fixture(scope="class")
    def btx_ftpz(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.properties = BTXParameterBlock(default={"valid_phase":
                                                     ('Liq', 'Vap'),
                                                     "activity_coeff_model":
                                                     "Ideal"})

        m.fs.unit = TrayColumn(default={
                               "number_of_trays": 10,
                               "feed_tray_location": 5,
                               "condenser_type": CondenserType.totalCondenser,
                               "condenser_temperature_spec":
                                   TemperatureSpec.atBubblePoint,
                               "property_package": m.fs.properties,
                               "has_heat_transfer": False,
                               "has_pressure_change": False})

        # Inlet feed conditions
        m.fs.unit.feed.flow_mol.fix(40)
        m.fs.unit.feed.temperature.fix(368)
        m.fs.unit.feed.pressure.fix(101325)
        m.fs.unit.feed.mole_frac_comp[0, "benzene"].fix(0.5)
        m.fs.unit.feed.mole_frac_comp[0, "toluene"].fix(0.5)

        # unit level inputs
        m.fs.unit.condenser.reflux_ratio.fix(1.4)
        m.fs.unit.condenser.condenser_pressure.fix(101325)

        m.fs.unit.reboiler.boilup_ratio.fix(1.3)

        return m

    @pytest.fixture(scope="class")
    def btx_fctp(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.properties = BTXParameterBlock(default={"valid_phase":
                                                     ('Liq', 'Vap'),
                                                     "activity_coeff_model":
                                                     "Ideal",
                                                     "state_vars": "FcTP"})

        m.fs.unit = TrayColumn(default={
                               "number_of_trays": 10,
                               "feed_tray_location": 5,
                               "condenser_type": CondenserType.totalCondenser,
                               "condenser_temperature_spec":
                                   TemperatureSpec.atBubblePoint,
                               "property_package": m.fs.properties,
                               "has_heat_transfer": False,
                               "has_pressure_change": False})

        # Inlet feed conditions
        m.fs.unit.feed.flow_mol_comp[0, "benzene"].fix(20)
        m.fs.unit.feed.flow_mol_comp[0, "toluene"].fix(20)
        m.fs.unit.feed.temperature.fix(368)
        m.fs.unit.feed.pressure.fix(101325)

        # unit level inputs
        m.fs.unit.condenser.reflux_ratio.fix(1.4)
        m.fs.unit.condenser.condenser_pressure.fix(101325)

        m.fs.unit.reboiler.boilup_ratio.fix(1.3)

        return m

    @pytest.mark.unit
    def test_build(self, btx_ftpz, btx_fctp):
        assert len(btx_ftpz.fs.unit.config) == 12

        assert btx_ftpz.fs.unit.feed_tray.config.is_feed_tray

        assert hasattr(btx_ftpz.fs.unit, "condenser")
        assert hasattr(btx_ftpz.fs.unit, "reboiler")

        assert hasattr(btx_ftpz.fs.unit, "rectification_section")
        assert hasattr(btx_ftpz.fs.unit, "stripping_section")

        assert len(btx_fctp.fs.unit.config) == 12

        assert btx_fctp.fs.unit.feed_tray.config.is_feed_tray

        assert hasattr(btx_fctp.fs.unit, "condenser")
        assert hasattr(btx_fctp.fs.unit, "reboiler")

        assert hasattr(btx_fctp.fs.unit, "rectification_section")
        assert hasattr(btx_fctp.fs.unit, "stripping_section")

    @pytest.mark.unit
    def test_dof(self, btx_ftpz, btx_fctp):
        assert degrees_of_freedom(btx_ftpz.fs.unit) == 0
        assert degrees_of_freedom(btx_fctp.fs.unit) == 0

    @pytest.mark.component
    def test_units_FTPz(self, btx_ftpz, btx_fctp):
        assert_units_consistent(btx_ftpz)

    @pytest.mark.component
    def test_units_FcTP(self, btx_fctp):
        assert_units_consistent(btx_fctp)

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, btx_ftpz, btx_fctp):
        initialization_tester(btx_ftpz)
        initialization_tester(btx_fctp)

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, btx_ftpz, btx_fctp):
        results = solver.solve(btx_ftpz)

        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

        results = solver.solve(btx_fctp)

        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, btx_ftpz, btx_fctp):

        # Distillate port - btx_ftpz
        assert (pytest.approx(18.978, rel=1e-2) ==
                value(btx_ftpz.fs.unit.condenser.distillate.flow_mol[0]))
        assert (pytest.approx(0.8882, rel=1e-2) ==
                value(btx_ftpz.fs.unit.condenser.
                      distillate.mole_frac_comp[0, "benzene"]))
        assert (pytest.approx(0.1118, rel=1e-2) ==
                value(btx_ftpz.fs.unit.condenser.
                      distillate.mole_frac_comp[0, "toluene"]))
        assert (pytest.approx(355.642, abs=1e-2) ==
                value(btx_ftpz.fs.unit.condenser.
                      distillate.temperature[0]))
        assert (pytest.approx(101325, abs=1e-3) ==
                value(btx_ftpz.fs.unit.condenser.
                      distillate.pressure[0]))

        # Bottoms port - btx_ftpz
        assert (pytest.approx(21.021, rel=1e-2) ==
                value(btx_ftpz.fs.unit.reboiler.bottoms.flow_mol[0]))
        assert (pytest.approx(0.1495, rel=1e-2) ==
                value(btx_ftpz.fs.unit.reboiler.
                      bottoms.mole_frac_comp[0, "benzene"]))
        assert (pytest.approx(0.8504, rel=1e-3) ==
                value(btx_ftpz.fs.unit.reboiler.
                      bottoms.mole_frac_comp[0, "toluene"]))
        assert (pytest.approx(377.337, abs=1e-2) ==
                value(btx_ftpz.fs.unit.reboiler.
                      bottoms.temperature[0]))
        assert (pytest.approx(101325, abs=1e-3) ==
                value(btx_ftpz.fs.unit.reboiler.
                      bottoms.pressure[0]))

        # Distillate port - btx_fctp
        assert (pytest.approx(16.856, rel=1e-2) ==
                value(btx_fctp.fs.unit.condenser.
                      distillate.flow_mol_comp[0, "benzene"]))
        assert (pytest.approx(2.121, rel=1e-2) ==
                value(btx_fctp.fs.unit.condenser.
                      distillate.flow_mol_comp[0, "toluene"]))
        assert (pytest.approx(355.638, abs=1e-2) ==
                value(btx_fctp.fs.unit.condenser.
                      distillate.temperature[0]))
        assert (pytest.approx(101325, abs=1e-3) ==
                value(btx_fctp.fs.unit.condenser.
                      distillate.pressure[0]))

        # # Bottoms port - btx_fctp
        assert (pytest.approx(3.179, rel=1e-2) ==
                value(btx_fctp.fs.unit.reboiler.
                      bottoms.flow_mol_comp[0, "benzene"]))
        assert (pytest.approx(17.876, rel=1e-3) ==
                value(btx_fctp.fs.unit.reboiler.
                      bottoms.flow_mol_comp[0, "toluene"]))
        assert (pytest.approx(377.28, abs=1e-2) ==
                value(btx_fctp.fs.unit.reboiler.
                      bottoms.temperature[0]))
        assert (pytest.approx(101325, abs=1e-3) ==
                value(btx_fctp.fs.unit.reboiler.
                      bottoms.pressure[0]))


class TestBTXIdealGeneric():
    @pytest.fixture(scope="class")
    def btx_ftpz_generic(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.properties = GenericParameterBlock(default=configuration)

        m.fs.unit = TrayColumn(default={
                               "number_of_trays": 10,
                               "feed_tray_location": 5,
                               "condenser_type": CondenserType.totalCondenser,
                               "condenser_temperature_spec":
                                   TemperatureSpec.atBubblePoint,
                               "property_package": m.fs.properties,
                               "has_heat_transfer": False,
                               "has_pressure_change": False})

        # Inlet feed conditions
        m.fs.unit.feed.flow_mol.fix(100)
        m.fs.unit.feed.temperature.fix(368)
        m.fs.unit.feed.pressure.fix(101325)
        m.fs.unit.feed.mole_frac_comp[0, "benzene"].fix(0.5)
        m.fs.unit.feed.mole_frac_comp[0, "toluene"].fix(0.5)

        # unit level inputs
        m.fs.unit.condenser.reflux_ratio.fix(1.4)
        m.fs.unit.condenser.condenser_pressure.fix(101325)

        m.fs.unit.reboiler.boilup_ratio.fix(1.3)

        return m

    @pytest.mark.unit
    def test_build(self, btx_ftpz_generic):
        assert len(btx_ftpz_generic.fs.unit.config) == 12

        assert btx_ftpz_generic.fs.unit.feed_tray.config.is_feed_tray

        assert hasattr(btx_ftpz_generic.fs.unit, "condenser")
        assert hasattr(btx_ftpz_generic.fs.unit, "reboiler")

        assert hasattr(btx_ftpz_generic.fs.unit, "rectification_section")
        assert hasattr(btx_ftpz_generic.fs.unit, "stripping_section")

        assert len(btx_ftpz_generic.fs.unit.config) == 12

    @pytest.mark.unit
    def test_dof(self, btx_ftpz_generic):
        assert degrees_of_freedom(btx_ftpz_generic) == 0

    @pytest.mark.component
    def test_units_FTPz(self, btx_ftpz_generic):
        assert_units_consistent(btx_ftpz_generic)

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, btx_ftpz_generic):
        initialization_tester(btx_ftpz_generic)

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, btx_ftpz_generic):
        results = solver.solve(btx_ftpz_generic)

        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, btx_ftpz_generic):

        # Distillate port - btx_ftpz
        assert (pytest.approx(49.28, rel=1e-2) ==
                value(
                    btx_ftpz_generic.fs.unit.condenser.distillate.flow_mol[0]))
        assert (pytest.approx(0.8784, rel=1e-2) ==
                value(btx_ftpz_generic.fs.unit.condenser.
                      distillate.mole_frac_comp[0, "benzene"]))
        assert (pytest.approx(0.1215, rel=1e-2) ==
                value(btx_ftpz_generic.fs.unit.condenser.
                      distillate.mole_frac_comp[0, "toluene"]))
        assert (pytest.approx(355.853, abs=1e-2) ==
                value(btx_ftpz_generic.fs.unit.condenser.
                      distillate.temperature[0]))
        assert (pytest.approx(101325, abs=1e-3) ==
                value(btx_ftpz_generic.fs.unit.condenser.
                      distillate.pressure[0]))

        # Bottoms port - btx_ftpz
        assert (pytest.approx(50.711, rel=1e-2) ==
                value(btx_ftpz_generic.fs.unit.reboiler.bottoms.flow_mol[0]))
        assert (pytest.approx(0.1322, rel=1e-2) ==
                value(btx_ftpz_generic.fs.unit.reboiler.
                      bottoms.mole_frac_comp[0, "benzene"]))
        assert (pytest.approx(0.8677, rel=1e-3) ==
                value(btx_ftpz_generic.fs.unit.reboiler.
                      bottoms.mole_frac_comp[0, "toluene"]))
        assert (pytest.approx(378.0433, abs=1e-2) ==
                value(btx_ftpz_generic.fs.unit.reboiler.
                      bottoms.temperature[0]))
        assert (pytest.approx(101325, abs=1e-3) ==
                value(btx_ftpz_generic.fs.unit.reboiler.
                      bottoms.pressure[0]))



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
3 stream IDAES heat exchanger model with given UA.
side 1 is hot stream, side 2 and 3 are cold streams
"""
# Import Pyomo libraries
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (ControlVolume0DBlock,
                        declare_process_block_class,
                        MaterialBalanceType,
                        EnergyBalanceType,
                        MomentumBalanceType,
                        UnitModelBlockData,
                        useDefault)

from idaes.core.util.config import is_physical_parameter_block
import idaes.core.util.scaling as iscale
from idaes.core.util import get_solver

import idaes.logger as idaeslog


# Additional import for the unit operation
from pyomo.environ import SolverFactory, Var, Reference

__author__ = "Boiler Subsystem Team (J. Ma, M. Zamarripa)"
__version__ = "1.0.0"


@declare_process_block_class("HeatExchangerWith3Streams")
class HeatExchangerWith3StreamsData(UnitModelBlockData):
    """
    Standard Heat Exchanger Unit Model Class
    """
    CONFIG = UnitModelBlockData.CONFIG()
    CONFIG.declare("side_1_property_package", ConfigValue(
        default=useDefault,
        domain=is_physical_parameter_block,
        description="Property package to use for control volume",
        doc="""Property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PhysicalParameterObject** - a PhysicalParameterBlock object.}"""))
    CONFIG.declare("side_1_property_package_args", ConfigBlock(
        implicit=True,
        description="Arguments to use for constructing property packages",
        doc="""A ConfigBlock with arguments to be passed to a property block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}"""))
    CONFIG.declare("side_2_property_package", ConfigValue(
        default=useDefault,
        domain=is_physical_parameter_block,
        description="Property package to use for control volume",
        doc="""Property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PhysicalParameterObject** - a PhysicalParameterBlock object.}"""))
    CONFIG.declare("side_2_property_package_args", ConfigBlock(
        implicit=True,
        description="Arguments to use for constructing property packages",
        doc="""A ConfigBlock with arguments to be passed to a property block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}"""))
    CONFIG.declare("side_3_property_package", ConfigValue(
        default=useDefault,
        domain=is_physical_parameter_block,
        description="Property package to use for control volume",
        doc="""Property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PhysicalParameterObject** - a PhysicalParameterBlock object.}"""))
    CONFIG.declare("side_3_property_package_args", ConfigBlock(
        implicit=True,
        description="Arguments to use for constructing property packages",
        doc="""A ConfigBlock with arguments to be passed to a property block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}"""))
    CONFIG.declare("material_balance_type", ConfigValue(
        default=MaterialBalanceType.componentPhase,
        domain=In(MaterialBalanceType),
        description="Material balance construction flag",
        doc="""Indicates what type of material balance should be constructed,
**default** - MaterialBalanceType.componentPhase.
**Valid values:** {
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}"""))
    CONFIG.declare("energy_balance_type", ConfigValue(
        default=EnergyBalanceType.enthalpyTotal,
        domain=In(EnergyBalanceType),
        description="Energy balance construction flag",
        doc="""Indicates what type of energy balance should be constructed,
**default** - EnergyBalanceType.enthalpyTotal.
**Valid values:** {
**EnergyBalanceType.none** - exclude energy balances,
**EnergyBalanceType.enthalpyTotal** - single ethalpy balance for material,
**EnergyBalanceType.enthalpyPhase** - ethalpy balances for each phase,
**EnergyBalanceType.energyTotal** - single energy balance for material,
**EnergyBalanceType.energyPhase** - energy balances for each phase.}"""))
    CONFIG.declare("momentum_balance_type", ConfigValue(
        default=MomentumBalanceType.pressureTotal,
        domain=In(MomentumBalanceType),
        description="Momentum balance construction flag",
        doc="""Indicates what type of momentum balance should be constructed,
**default** - MomentumBalanceType.pressureTotal.
**Valid values:** {
**MomentumBalanceType.none** - exclude momentum balances,
**MomentumBalanceType.pressureTotal** - single pressure balance for material,
**MomentumBalanceType.pressurePhase** - pressure balances for each phase,
**MomentumBalanceType.momentumTotal** - single momentum balance for material,
**MomentumBalanceType.momentumPhase** - momentum balances for each phase.}"""))
    CONFIG.declare("has_heat_transfer", ConfigValue(
        default=True,
        domain=In([True, False]),
        description="Heat transfer term construction flag",
        doc="""Indicates whether terms for heat transfer should be constructed,
**default** - False.
**Valid values:** {
**True** - include heat transfer terms,
**False** - exclude heat transfer terms.}"""))
    CONFIG.declare("has_pressure_change", ConfigValue(
        default=False,
        domain=In([True, False]),
        description="Pressure change term construction flag",
        doc="""Indicates whether terms for pressure change should be
constructed,
**default** - False.
**Valid values:** {
**True** - include pressure change terms,
**False** - exclude pressure change terms.}"""))
    CONFIG.declare("flow_type_side_2", ConfigValue(
        default='counter-current',
        domain=In(['counter-current', 'co-current']),
        description="Flow configuration in unit",
        doc="""Flag indicating type of flow arrangement to use for heat
exchanger, **default** 'counter-current' counter-current flow arrangement"""))
    CONFIG.declare("flow_type_side_3", ConfigValue(
        default='counter-current',
        domain=In(['counter-current', 'co-current']),
        description="Flow configuration in unit",
        doc="""Flag indicating type of flow arrangement to use for heat
exchanger (default = 'counter-current' - counter-current flow arrangement"""))

    def build(self):
        """
        Begin building model
        """
        # Call UnitModel.build to setup dynamics
        super(HeatExchangerWith3StreamsData, self).build()

        # Build Holdup Block
        self.side_1 = ControlVolume0DBlock(default={
            "dynamic": self.config.dynamic,
            "has_holdup": self.config.has_holdup,
            "property_package": self.config.side_1_property_package,
            "property_package_args": self.config.side_1_property_package_args})

        self.side_2 = ControlVolume0DBlock(default={
            "dynamic": self.config.dynamic,
            "has_holdup": self.config.has_holdup,
            "property_package": self.config.side_2_property_package,
            "property_package_args": self.config.side_2_property_package_args})

        self.side_3 = ControlVolume0DBlock(default={
            "dynamic": self.config.dynamic,
            "has_holdup": self.config.has_holdup,
            "property_package": self.config.side_3_property_package,
            "property_package_args": self.config.side_3_property_package_args})

        # Add Geometry
        self.side_1.add_geometry()
        self.side_2.add_geometry()
        self.side_3.add_geometry()

        # Add state block
        self.side_1.add_state_blocks(has_phase_equilibrium=False)

        # Add material balance
        self.side_1.add_material_balances(
            balance_type=self.config.material_balance_type)
        # add energy balance
        self.side_1.add_energy_balances(
            balance_type=self.config.energy_balance_type,
            has_heat_transfer=self.config.has_heat_transfer)
        # add momentum balance
        self.side_1.add_momentum_balances(
            balance_type=self.config.momentum_balance_type,
            has_pressure_change=self.config.has_pressure_change)

        # Add state block
        self.side_2.add_state_blocks(has_phase_equilibrium=False)

        # Add material balance
        self.side_2.add_material_balances(
            balance_type=self.config.material_balance_type)
        # add energy balance
        self.side_2.add_energy_balances(
            balance_type=self.config.energy_balance_type,
            has_heat_transfer=self.config.has_heat_transfer)
        # add momentum balance
        self.side_2.add_momentum_balances(
            balance_type=self.config.momentum_balance_type,
            has_pressure_change=self.config.has_pressure_change)

        # Add state block
        self.side_3.add_state_blocks(has_phase_equilibrium=False)

        # Add material balance
        self.side_3.add_material_balances(
            balance_type=self.config.material_balance_type)
        # add energy balance
        self.side_3.add_energy_balances(
            balance_type=self.config.energy_balance_type,
            has_heat_transfer=self.config.has_heat_transfer)
        # add momentum balance
        self.side_3.add_momentum_balances(
            balance_type=self.config.momentum_balance_type,
            has_pressure_change=self.config.has_pressure_change)

        self._set_geometry()

        # Construct performance equations
        self._make_performance()

        # Construct performance equations
        if self.config.flow_type_side_2 == "counter-current":
            self._make_counter_current_side_2()
        else:
            self._make_co_current_side_2()

        # Construct performance equations
        if self.config.flow_type_side_3 == "counter-current":
            self._make_counter_current_side_3()
        else:
            self._make_co_current_side_3()

        self.add_inlet_port(name="side_1_inlet", block=self.side_1)
        self.add_inlet_port(name="side_2_inlet", block=self.side_2)
        self.add_inlet_port(name="side_3_inlet", block=self.side_3)
        self.add_outlet_port(name="side_1_outlet", block=self.side_1)
        self.add_outlet_port(name="side_2_outlet", block=self.side_2)
        self.add_outlet_port(name="side_3_outlet", block=self.side_3)

    def _set_geometry(self):
        """
        Define the geometry of the unit as necessary, and link to holdup volume

        """

        # UA (product of overall heat transfer coefficient and area)
        # between side 1 and side 2
        self.ua_side_2 = Var(self.flowsheet().config.time, initialize=10.0,
                             doc='UA between side 1 and side 2')

        # UA (product of overall heat transfer coefficient and area)
        # between side 1 and side 3
        self.ua_side_3 = Var(self.flowsheet().config.time, initialize=10.0,
                             doc='UA between side 1 and side 3')

        # fraction of heat from hot stream as heat loss to ambient
        self.frac_heatloss = Var(initialize=0.05,
                                 doc='Fraction of heat loss to ambient')

        if self.config.has_holdup is True:
            self.volume_side_1 = Reference(self.side_1.volume)
            self.volume_side_2 = Reference(self.side_2.volume)
            self.volume_side_3 = Reference(self.side_3.volume)

    def _make_performance(self):
        """
        Define constraints which describe the behaviour of the unit model.

        Args:
            None

        Returns:
            None
        """
        # Set references to balance terms at unit level
        self.heat_duty_side_1 = Reference(self.side_1.heat)
        self.heat_duty_side_2 = Reference(self.side_2.heat)
        self.heat_duty_side_3 = Reference(self.side_3.heat)

        if self.config.has_pressure_change is True:
            self.deltaP_side_1 = Reference(self.side_1.deltaP)
            self.deltaP_side_2 = Reference(self.side_2.deltaP)
            self.deltaP_side_3 = Reference(self.side_3.deltaP)

        # Performance parameters and variables
        # Temperature driving force
        self.temperature_driving_force_side_2 = Var(self.flowsheet().config.
                                                    time,
                                                    initialize=1.0,
                                                    doc='Mean driving force '
                                                    'for heat exchange')

        # Temperature driving force
        self.temperature_driving_force_side_3 = Var(self.flowsheet().
                                                    config.time,
                                                    initialize=1.0,
                                                    doc='Mean driving force '
                                                    'for heat exchange')

        # Temperature difference at side 2 inlet
        self.side_2_inlet_dT = Var(self.flowsheet().config.time,
                                   initialize=1.0,
                                   doc='Temperature difference '
                                   'at side 2 inlet')

        # Temperature difference at side 2 outlet
        self.side_2_outlet_dT = Var(self.flowsheet().config.time,
                                    initialize=1.0,
                                    doc='Temperature difference '
                                    'at side 2 outlet')

        # Temperature difference at side 3 inlet
        self.side_3_inlet_dT = Var(self.flowsheet().config.time,
                                   initialize=1.0,
                                   doc='Temperature difference'
                                   ' at side 3 inlet')

        # Temperature difference at side 3 outlet
        self.side_3_outlet_dT = Var(self.flowsheet().config.time,
                                    initialize=1.0,
                                    doc='Temperature difference '
                                    'at side 3 outlet')

        # Driving force side 2 (Underwood approximation)
        @self.Constraint(self.flowsheet().config.time,
                         doc="Log mean temperature difference calculation "
                         "using Underwood approximation")
        def LMTD_side_2(b, t):
            return b.temperature_driving_force_side_2[t] == \
                ((b.side_2_inlet_dT[t]**(1/3)
                  + b.side_2_outlet_dT[t]**(1/3))/2)**(3)

        # Driving force side 3 (Underwood approximation)
        @self.Constraint(self.flowsheet().config.time,
                         doc="Log mean temperature difference calculation "
                         "using Underwood approximation")
        def LMTD_side_3(b, t):
            return b.temperature_driving_force_side_3[t] == \
                ((b.side_3_inlet_dT[t]**(1/3)
                  + b.side_3_outlet_dT[t]**(1/3))/2)**(3)

        # Heat duty side 2
        @self.Constraint(self.flowsheet().config.time,
                         doc="Heat transfer rate")
        def heat_duty_side_2_eqn(b, t):
            return b.heat_duty_side_2[t] == \
                (b.ua_side_2[t] * b.temperature_driving_force_side_2[t])

        # Heat duty side 3
        @self.Constraint(self.flowsheet().config.time,
                         doc="Heat transfer rate")
        def heat_duty_side_3_eqn(b, t):
            return b.heat_duty_side_3[t] == \
                (b.ua_side_3[t]*b.temperature_driving_force_side_3[t])

        # Energy balance equation
        @self.Constraint(self.flowsheet().config.time,
                         doc="Energy balance between two sides")
        def heat_duty_side_1_eqn(b, t):
            return -b.heat_duty_side_1[t]*(1-b.frac_heatloss) == \
                (b.heat_duty_side_2[t] + b.heat_duty_side_3[t])

    def _make_co_current_side_2(self):
        """
        Add temperature driving force Constraints for co-current flow.


        """
        # Temperature Differences
        @self.Constraint(self.flowsheet().config.time,
                         doc="Side 2 inlet temperature difference")
        def side_2_inlet_dT_eqn(b, t):
            return b.side_2_inlet_dT[t] == (
                       b.side_1.properties_in[t].temperature -
                       b.side_2.properties_in[t].temperature)

        @self.Constraint(self.flowsheet().config.time,
                         doc="Side 2 outlet temperature difference")
        def side_2_outlet_dT_eqn(b, t):
            return b.side_2_outlet_dT[t] == (
                       b.side_1.properties_out[t].temperature -
                       b.side_2.properties_out[t].temperature)

    def _make_counter_current_side_2(self):
        """
        Add temperature driving force Constraints for counter-current flow.
        """
        # Temperature Differences
        @self.Constraint(self.flowsheet().config.time,
                         doc="Side 2 inlet temperature difference")
        def side_2_inlet_dT_eqn(b, t):
            return b.side_2_inlet_dT[t] == (
                       b.side_1.properties_out[t].temperature -
                       b.side_2.properties_in[t].temperature)

        @self.Constraint(self.flowsheet().config.time,
                         doc="Side 2 outlet temperature difference")
        def side_2_outlet_dT_eqn(b, t):
            return b.side_2_outlet_dT[t] == (
                       b.side_1.properties_in[t].temperature -
                       b.side_2.properties_out[t].temperature)

    def _make_co_current_side_3(self):
        """
        Add temperature driving force Constraints for co-current flow.
        """
        # Temperature Differences
        @self.Constraint(self.flowsheet().config.time,
                         doc="Side 3 inlet temperature difference")
        def side_3_inlet_dT_eqn(b, t):
            return b.side_3_inlet_dT[t] == (
                       b.side_1.properties_in[t].temperature -
                       b.side_3.properties_in[t].temperature)

        @self.Constraint(self.flowsheet().config.time,
                         doc="Side 3 outlet temperature difference")
        def side_3_outlet_dT_eqn(b, t):
            return b.side_3_outlet_dT[t] == (
                       b.side_1.properties_out[t].temperature -
                       b.side_3.properties_out[t].temperature)

    def _make_counter_current_side_3(self):
        """
        Add temperature driving force Constraints for counter-current flow.
        """
        # Temperature Differences
        @self.Constraint(self.flowsheet().config.time,
                         doc="Side 3 inlet temperature difference")
        def side_3_inlet_dT_eqn(b, t):
            return b.side_3_inlet_dT[t] == (
                       b.side_1.properties_out[t].temperature -
                       b.side_3.properties_in[t].temperature)

        @self.Constraint(self.flowsheet().config.time,
                         doc="Side 3 outlet temperature difference")
        def side_3_outlet_dT_eqn(b, t):
            return b.side_3_outlet_dT[t] == (
                       b.side_1.properties_in[t].temperature -
                       b.side_3.properties_out[t].temperature)

    def initialize(blk, state_args_1=None, state_args_2=None,
                   state_args_3=None, outlvl=idaeslog.NOTSET,
                   solver=None, optarg=None):
        '''
        General Heat Exchanger initialisation routine.

        Keyword Arguments:
            state_args_1 : a dict of arguments to be passed to the property
                           package(s) for side 1 of the heat exchanger to
                           provide an initial state for initialization
                           (see documentation of the specific property package)
                           (default = None).
            state_args_2 : a dict of arguments to be passed to the property
                           package(s) for side 2 of the heat exchanger to
                           provide an initial state for initialization
                           (see documentation of the specific property package)
                           (default = None).
            state_args_3 : a dict of arguments to be passed to the property
                           package(s) for side 3 of the heat exchanger to
                           provide an initial state for initialization
                           (see documentation of the specific property package)
                           (default = None).
            outlvl : sets output level of initialisation routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)

        Returns:
            None
        '''
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")

        # Create solver
        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # Initialize inlet property blocks
        flags1 = blk.side_1.initialize(outlvl=outlvl,
                                       optarg=optarg,
                                       solver=solver,
                                       state_args=state_args_1)

        flags2 = blk.side_2.initialize(outlvl=outlvl,
                                       optarg=optarg,
                                       solver=solver,
                                       state_args=state_args_2)

        flags3 = blk.side_3.initialize(outlvl=outlvl,
                                       optarg=optarg,
                                       solver=solver,
                                       state_args=state_args_3)

        init_log.info('Initialisation Step 1 Complete.')

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)
        init_log.info("Initialization Step 2 Complete: {}".format(
                idaeslog.condition(res)
            )
        )
        # ---------------------------------------------------------------------
        # Release Inlet state
        blk.side_1.release_state(flags1, outlvl)
        blk.side_2.release_state(flags2, outlvl)
        blk.side_3.release_state(flags3, outlvl)

        init_log.info_low("Initialization Complete: {}".format(
                idaeslog.condition(res)
            )
        )

    def calculate_scaling_factors(self):
        for t, c in self.heat_duty_side_1_eqn.items():
            sf = iscale.get_scaling_factor(
                self.heat_duty_side_1[t], default=1, warning=True)
            iscale.constraint_scaling_transform(c, sf, overwrite=False)

        for t, c in self.heat_duty_side_2_eqn.items():
            sf = iscale.get_scaling_factor(
                self.heat_duty_side_2[t], default=1, warning=True)
            iscale.constraint_scaling_transform(c, sf, overwrite=False)

        for t, c in self.heat_duty_side_3_eqn.items():
            sf = iscale.get_scaling_factor(
                self.heat_duty_side_3[t], default=1, warning=True)
            iscale.constraint_scaling_transform(c, sf, overwrite=False)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Property package for MEA Carbon Capture System Condenser

CO2-H2O VLE system
Ideal VLE with non-condensables
"""
# Import Python libraries
import logging

from pyomo.environ import units as pyunits

# Import IDAES cores
from idaes.core import LiquidPhase, VaporPhase, Component
from idaes.core.phases import PhaseType as PT

from idaes.generic_models.properties.core.state_definitions import FTPx
from idaes.generic_models.properties.core.eos.ideal import Ideal
from idaes.generic_models.properties.core.phase_equil import SmoothVLE
from idaes.generic_models.properties.core.phase_equil.bubble_dew import \
        IdealBubbleDew
from idaes.generic_models.properties.core.phase_equil.forms import fugacity

from idaes.generic_models.properties.core.pure.Perrys import Perrys
from idaes.generic_models.properties.core.pure.NIST import NIST

# Set up logger
_log = logging.getLogger(__name__)


# ---------------------------------------------------------------------
# Configuration dictionary for an ideal CO2-H2O system
# Assumptions
# 1. Ideal gas-liquid system
# 2. CO2 conc. negligible in liquid phase, at equilibrium
# 3. Properties are applicable at standard pressure (1 atm)

# Reference Temperature : 298.15 K
# Reference Pressure : 101325 Pa   

# Data Sources:

# [1] NIST Webbook, https://webbook.nist.gov/
#     Retrieved 27th November, 2020. Converted from bar to Pa
# [2] Perry's Chemical Engineers' Handbook 7th Ed.

# Initial temperature and pressure reference:
# Measurement and Modeling of the Phase Behavior of the 
# (Carbon Dioxide + Water) Mixture at Temperatures From 298.15 K to 448.15 
# KShu-Xin Hou, Geoffrey C. Maitland, and J. P. Martin Trusler*
# Qatar Carbonates and Carbon Storage Research Centre 
# Department of Chemical Engineering Imperial College London, 
# South Kensington Campus, London SW7 2AZ.  U.K

configuration = {
    # Specifying components
    "components": {
        'H2O': {"type": Component,
                "dens_mol_liq_comp": Perrys,
                "enth_mol_liq_comp": Perrys,
                "enth_mol_ig_comp": NIST,
                "pressure_sat_comp": NIST,
                "phase_equilibrium_form": {("Vap", "Liq"): fugacity},
                "parameter_data": {
                    "mw": (18.0153E-3, pyunits.kg/pyunits.mol),  # [1]
                    "pressure_crit": (220.64E5, pyunits.Pa),  # [1]
                    "temperature_crit": (647, pyunits.K),  # [1]
                    "dens_mol_liq_comp_coeff": {
                        '1': (5.459, pyunits.kmol*pyunits.m**-3),  # [2] pg. 2-98, temperature range 273.16 K - 333.15 K
                        '2': (0.30542, None),
                        '3': (647.13, pyunits.K),
                        '4': (0.081, None)},
                    "cp_mol_ig_comp_coeff": {
                        'A': (30.09200, pyunits.J/pyunits.mol/pyunits.K),  # [1] temperature range 500 K- 1700 K
                        'B': (6.832514, pyunits.J*pyunits.mol**-1*pyunits.K**-1*pyunits.kiloK**-1),
                        'C': (6.793435, pyunits.J*pyunits.mol**-1*pyunits.K**-1*pyunits.kiloK**-2),
                        'D': (-2.534480, pyunits.J*pyunits.mol**-1*pyunits.K**-1*pyunits.kiloK**-3),
                        'E': (0.082139, pyunits.J*pyunits.mol**-1*pyunits.K**-1*pyunits.kiloK**2),
                        'F': (-250.8810, pyunits.kJ/pyunits.mol),
                        'G': (223.3967, pyunits.J/pyunits.mol/pyunits.K),
                        'H': (0, pyunits.kJ/pyunits.mol)},
                    "cp_mol_liq_comp_coeff": {
                        '1': (2.7637E5, pyunits.J/pyunits.kmol/pyunits.K),  # [2] pg 2-174, temperature range 273.16 K - 533.15 K
                        '2': (-2.0901E3, pyunits.J/pyunits.kmol/pyunits.K**2),
                        '3': (8.125, pyunits.J/pyunits.kmol/pyunits.K**3),
                        '4': (-1.4116E-2, pyunits.J/pyunits.kmol/pyunits.K**4),
                        '5': (9.3701E-6, pyunits.J/pyunits.kmol/pyunits.K**5)},
                    "enth_mol_form_liq_comp_ref": (
                        -285.83E3, pyunits.J/pyunits.mol),  # [1]
                    "enth_mol_form_vap_comp_ref": (
                        0, pyunits.J/pyunits.mol),  # [1]
                    "pressure_sat_comp_coeff": {
                        'A': (4.6543, None),  # [1], temperature range 255.9 K - 373 K
                        'B': (1435.264, pyunits.K),
                        'C': (-64.848, pyunits.K)}}},
        'CO2': {"type": Component,
                "valid_phase_types": PT.vaporPhase,
                "enth_mol_ig_comp": NIST,
                "parameter_data": {
                   "mw": (44.0095E-3, pyunits.kg/pyunits.mol),  # [1]
                   "pressure_crit": (73.825E5, pyunits.Pa),  # [1]
                   "temperature_crit": (304.23, pyunits.K),  # [1]
                   "cp_mol_ig_comp_coeff": {                 # [1], temperature range 298 K - 1200 K
                       'A': (24.99735, pyunits.J/pyunits.mol/pyunits.K),
                       'B': (55.18696, pyunits.J*pyunits.mol**-1*pyunits.K**-1*pyunits.kiloK**-1),
                       'C': (-33.69137, pyunits.J*pyunits.mol**-1*pyunits.K**-1*pyunits.kiloK**-2),
                       'D': (7.948387, pyunits.J*pyunits.mol**-1*pyunits.K**-1*pyunits.kiloK**-3),
                       'E': (-0.136638, pyunits.J*pyunits.mol**-1*pyunits.K**-1*pyunits.kiloK**2),
                       'F': (-403.6075, pyunits.kJ/pyunits.mol),
                       'G': (228.2431, pyunits.J/pyunits.mol/pyunits.K),
                       'H': (0, pyunits.kJ/pyunits.mol)},
                   "enth_mol_form_vap_comp_ref": (0, pyunits.J/pyunits.mol)# [1]
                         }}},

    # Specifying phases
    "phases":  {'Liq': {"type": LiquidPhase,
                        "equation_of_state": Ideal},
                'Vap': {"type": VaporPhase,
                        "equation_of_state": Ideal}},

    # Set base units of measurement
    "base_units": {"time": pyunits.s,
                   "length": pyunits.m,
                   "mass": pyunits.kg,
                   "amount": pyunits.mol,
                   "temperature": pyunits.K},

    # Specifying state definition
    "state_definition": FTPx,
    "state_bounds": {"flow_mol": (0, 10, 20, pyunits.mol/pyunits.s),
                     "temperature": (273.15, 323.15, 1000, pyunits.K),
                     "pressure": (5E4, 108900, 1e7, pyunits.Pa),
                     "mole_frac_comp": {"H2O":(0,0.5,1),"CO2":(0,0.5,1)}},
    "pressure_ref": (101325, pyunits.Pa),
    "temperature_ref": (298.15, pyunits.K),

    # Defining phase equilibria
    "phases_in_equilibrium": [("Vap", "Liq")],
    "phase_equilibrium_state": {("Vap", "Liq"): SmoothVLE},
    "bubble_dew_method": IdealBubbleDew}



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Example Peng-Robinson parameter block for the VLE calucations for a
benzene-toluene system.

Unless otherwise noted, parameters are from:
"The Properties of Gases and Liquids, 4th Edition", Reid, Prausnitz and Poling,
McGraw-Hill, 1987
"""
# Import Pyomo libraries
from pyomo.environ import Reals, Param, NonNegativeReals, Set, units as pyunits

# Import IDAES cores
from idaes.core import declare_process_block_class, Component
from idaes.core.util.misc import extract_data

from idaes.generic_models.properties.cubic_eos.cubic_prop_pack import (
        CubicParameterData, CubicEoS)
from idaes.logger import getIdaesLogger


# Set up logger
_log = getIdaesLogger(__name__)


@declare_process_block_class("BTParameterBlock")
class BTParameterData(CubicParameterData):

    def build(self):
        '''
        Callable method for Block construction.
        '''
        super(BTParameterData, self).build()

        self.cubic_type = CubicEoS.PR

        # Add Component objects
        self.benzene = Component(default={
            "elemental_composition": {"C": 6, "H": 6}})
        self.toluene = Component(default={
            "elemental_composition": {"C": 7, "H": 8}})

        # List of phase equilibrium index
        self.phase_equilibrium_idx = Set(initialize=[1, 2])

        self.phase_equilibrium_list = \
            {1: ["benzene", ("Vap", "Liq")],
             2: ["toluene", ("Vap", "Liq")]}

        # Thermodynamic reference state
        self.pressure_ref = Param(mutable=True,
                                  default=101325,
                                  doc='Reference pressure [Pa]',
                                  units=pyunits.Pa)
        self.temperature_ref = Param(mutable=True,
                                     default=298.15,
                                     doc='Reference temperature [K]',
                                     units=pyunits.K)

        # Critical Properties
        pressure_crit_data = {'benzene': 48.9e5,
                              'toluene': 41.0e5}

        self.pressure_crit = Param(
            self.component_list,
            within=NonNegativeReals,
            mutable=False,
            initialize=extract_data(pressure_crit_data),
            doc='Critical pressure [Pa]',
            units=pyunits.Pa)

        temperature_crit_data = {'benzene': 562.2,
                                 'toluene': 591.8}

        self.temperature_crit = Param(
            self.component_list,
            within=NonNegativeReals,
            mutable=False,
            initialize=extract_data(temperature_crit_data),
            doc='Critical temperature [K]',
            units=pyunits.K)

        # Pitzer acentricity factor
        omega_data = {'benzene': 0.212,
                      'toluene': 0.263}

        self.omega = Param(
            self.component_list,
            within=Reals,
            mutable=False,
            initialize=extract_data(omega_data),
            doc='Acentricity Factor')

        # Peng-Robinson binary interaction parameters
        kappa_data = {
            ('benzene', 'benzene'): 0.0000, ('benzene', 'toluene'): 0.0000,
            ('toluene', 'benzene'): 0.0000, ('toluene', 'toluene'): 0.0000}

        self.kappa = Param(
            self.component_list,
            self.component_list,
            within=Reals,
            mutable=False,
            initialize=extract_data(kappa_data),
            doc='Peng-Robinson binary interaction parameters')

        # Molecular Weights
        mw_comp_data = {'benzene': 78.1136E-3,
                        'toluene': 92.1405E-3}

        self.mw_comp = Param(self.component_list,
                             mutable=False,
                             initialize=extract_data(mw_comp_data),
                             doc="molecular weight kg/mol",
                             units=pyunits.kg/pyunits.mol)

        # Constants for specific heat capacity, enthalpy and entropy
        self.cp_mol_ig_comp_coeff_1 = Param(
            self.component_list,
            mutable=False,
            initialize={'benzene': -3.392E1,
                        'toluene': -2.435E1},
            doc="Parameter 1 to compute cp_mol_comp",
            units=pyunits.J/pyunits.mol/pyunits.K)

        self.cp_mol_ig_comp_coeff_2 = Param(
            self.component_list,
            mutable=False,
            initialize={'benzene': 4.739E-1,
                        'toluene': 5.125E-1},
            doc="Parameter 2 to compute cp_mol_comp",
            units=pyunits.J/pyunits.mol/pyunits.K**2)

        self.cp_mol_ig_comp_coeff_3 = Param(
            self.component_list,
            mutable=False,
            initialize={'benzene': -3.017E-4,
                        'toluene': -2.765E-4},
            doc="Parameter 3 to compute cp_mol_comp",
            units=pyunits.J/pyunits.mol/pyunits.K**3)

        self.cp_mol_ig_comp_coeff_4 = Param(
            self.component_list,
            mutable=False,
            initialize={'benzene': 7.130E-8,
                        'toluene': 4.911E-8},
            doc="Parameter 4 to compute cp_mol_comp",
            units=pyunits.J/pyunits.mol/pyunits.K**4)

        # Standard heats of formation
        # Source: NIST Webbook, https://webbook.nist.gov
        # Retrieved 25th September 2019
        dh_form_data = {'benzene': 82.9e3,
                        'toluene': 50.1e3}

        self.enth_mol_form_ref = Param(self.component_list,
                                       mutable=False,
                                       initialize=extract_data(dh_form_data),
                                       doc="Standard heats of formation",
                                       units=pyunits.J/pyunits.mol)

        # Standard entropy of formation
        # Source: Engineering Toolbox, https://www.engineeringtoolbox.com
        # Retrieved 25th September, 2019
        ds_form_data = {'benzene': -269,
                        'toluene': -321}

        self.entr_mol_form_ref = Param(self.component_list,
                                       mutable=False,
                                       initialize=extract_data(ds_form_data),
                                       doc="Standard entropy of formation",
                                       units=pyunits.J/pyunits.mol/pyunits.K)

        # Antoine coefficients for ideal vapour (units: bar, K)
        # This is needed for initial guesses of bubble and dew points
        self.antoine_coeff_A = Param(
            self.component_list,
            mutable=False,
            initialize={'benzene': 4.202,
                        'toluene': 4.216},
            doc="Antoine A Parameter to calculate pressure_sat",
            units=pyunits.dimensionless)

        self.antoine_coeff_B = Param(
            self.component_list,
            mutable=False,
            initialize={'benzene': 1322,
                        'toluene': 1435},
            doc="Antoine B Parameter to calculate pressure_sat",
            units=pyunits.K)

        self.antoine_coeff_C = Param(
            self.component_list,
            mutable=False,
            initialize={'benzene': -38.56,
                        'toluene': -43.33},
            doc="Antoine C Parameter to calculate pressure_sat",
            units=pyunits.K)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for math util methods.
"""

import pytest
from pyomo.environ import (Block, ConcreteModel, Constraint, Expression, exp,
                           Set, Var, value, Param, Reals,
                           TransformationFactory, TerminationCondition)
from pyomo.network import Arc, Port
from pyomo.core.base.units_container import UnitsError

from idaes.core import (FlowsheetBlock,
                        MaterialBalanceType,
                        EnergyBalanceType,
                        MomentumBalanceType,
                        declare_process_block_class,
                        PhysicalParameterBlock,
                        StateBlock,
                        StateBlockData,
                        ReactionParameterBlock,
                        ReactionBlockBase,
                        ReactionBlockDataBase,
                        MaterialFlowBasis)
from idaes.core.util.testing import PhysicalParameterTestBlock
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.generic_models.unit_models import CSTR
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.initialization import (fix_state_vars,
                                            revert_state_vars,
                                            propagate_state,
                                            solve_indexed_blocks,
                                            initialize_by_time_element)
from idaes.core.util import get_solver

__author__ = "Andrew Lee"


# Set up solver
solver = get_solver()


@declare_process_block_class("AqueousEnzymeParameterBlock")
class ParameterData(PhysicalParameterBlock):
    """
    Parameter block for the aqueous enzyme reaction in the biochemical CSTR
    used by Christofides and Daoutidis, 1996, presented by Heineken et al, 1967
    """

    def build(self):
        super(ParameterData, self).build()

        # all components are in the aqueous phase
        self.phase_list = Set(initialize=['aq'])
        self.component_list = Set(initialize=['S', 'E', 'C', 'P', 'Solvent'])

        self.state_block_class = AqueousEnzymeStateBlock

    @classmethod
    def define_metadata(cls, obj):
        obj.add_default_units({'time': 'min',
                               'length': 'm',
                               'amount': 'kmol',
                               'temperature': 'K',
                               'energy': 'kcal',
                               'holdup': 'kmol'})


class _AqueousEnzymeStateBlock(StateBlock):
    def initialize(blk):
        pass


@declare_process_block_class("AqueousEnzymeStateBlock",
                             block_class=_AqueousEnzymeStateBlock)
class AqueousEnzymeStateBlockData(StateBlockData):
    def build(self):
        super(AqueousEnzymeStateBlockData, self).build()

        self.conc_mol = Var(self._params.component_list,
                            domain=Reals,
                            doc='Component molar concentration [kmol/m^3]')

        self.flow_mol_comp = Var(self._params.component_list,
                                 domain=Reals,
                                 doc='Molar component flow rate [kmol/min]')

        self.flow_vol = Var(
            domain=Reals, doc='Volumetric flow rate out of reactor [m^3/min]')

        self.temperature = Var(initialize=303, domain=Reals,
                               doc='Temperature within reactor [K]')

        if not self.config.defined_state:
            self.conc_mol['Solvent'].fix(1.)

        def flow_mol_comp_rule(b, j):
            return b.flow_mol_comp[j] == b.flow_vol*b.conc_mol[j]

        self.flow_mol_comp_eqn = Constraint(
            self._params.component_list,
            rule=flow_mol_comp_rule,
            doc='Outlet component molar flow rate equation')

    def get_material_density_terms(b, p, j):
        return b.conc_mol[j]

    def get_material_flow_terms(b, p, j):
        return b.flow_mol_comp[j]

    def get_material_flow_basis(b):
        return MaterialFlowBasis.molar

    def get_enthalpy_flow_terms(b, p):
        return b.flow_vol*b.temperature

    def get_energy_density_terms(b, p):
        return b.temperature

    def define_state_vars(b):
        return {'conc_mol': b.conc_mol,
                'temperature': b.temperature,
                'flow_vol': b.flow_vol}


@declare_process_block_class('EnzymeReactionParameterBlock')
class EnzymeReactionParameterData(ReactionParameterBlock):
    '''
    Enzyme reaction:
    S + E <-> C -> P + E
    '''

    def build(self):
        super(EnzymeReactionParameterData, self).build()

        self.reaction_block_class = EnzymeReactionBlock

        self.rate_reaction_idx = Set(initialize=['R1', 'R2', 'R3'])
        self.rate_reaction_stoichiometry = {('R1', 'aq', 'S'): -1,
                                            ('R1', 'aq', 'E'): -1,
                                            ('R1', 'aq', 'C'): 1,
                                            ('R1', 'aq', 'P'): 0,
                                            ('R1', 'aq', 'Solvent'): 0,
                                            ('R2', 'aq', 'S'): 1,
                                            ('R2', 'aq', 'E'): 1,
                                            ('R2', 'aq', 'C'): -1,
                                            ('R2', 'aq', 'P'): 0,
                                            ('R2', 'aq', 'Solvent'): 0,
                                            ('R3', 'aq', 'S'): 0,
                                            ('R3', 'aq', 'E'): 1,
                                            ('R3', 'aq', 'C'): -1,
                                            ('R3', 'aq', 'P'): 1,
                                            ('R3', 'aq', 'Solvent'): 0}

        self.act_energy = Param(self.rate_reaction_idx,
                                initialize={'R1': 8.0e3,
                                            'R2': 9.0e3,
                                            'R3': 1.0e4},
                                doc='Activation energy [kcal/kmol]')

        self.gas_const = Param(initialize=1.987,
                               doc='Gas constant R [kcal/kmol/K]')

        self.temperature_ref = Param(initialize=300.0,
                                     doc='Reference temperature')

        self.k_rxn = Param(
            self.rate_reaction_idx,
            initialize={'R1': 3.36e6,
                        'R2': 1.80e6,
                        'R3': 5.79e7},
            doc='Pre-exponential rate constant in Arrhenius expression')

    #    self.reaction_block_class = EnzymeReactionBlock

    @classmethod
    def define_metadata(cls, obj):
        obj.add_default_units({'time': 'min',
                               'length': 'm',
                               'amount': 'kmol',
                               'temperature': 'K',
                               'energy': 'kcal',
                               'holdup': 'kmol'})


class _EnzymeReactionBlock(ReactionBlockBase):
    def initialize(blk):
        # initialize for reaction rates for each data object
        pass


@declare_process_block_class('EnzymeReactionBlock',
                             block_class=_EnzymeReactionBlock)
class EnzymeReactionBlockData(ReactionBlockDataBase):
    def build(self):
        super(EnzymeReactionBlockData, self).build()

        self.reaction_coef = Var(self._params.rate_reaction_idx,
                                 domain=Reals, doc='Reaction rate coefficient')

        self.reaction_rate = Var(self._params.rate_reaction_idx,
                                 domain=Reals,
                                 doc='Reaction rate [kmol/m^3/min]')

        self.dh_rxn = Param(self._params.rate_reaction_idx,
                            domain=Reals, doc='Heat of reaction',
                            initialize={'R1': 1e3/900/0.231,
                                        'R2': 1e3/900/0.231,
                                        'R3': 5e3/900/0.231})

        def reaction_rate_rule(b, r):
            if r == 'R1':
                return (b.reaction_rate[r] ==
                        b.reaction_coef[r] *
                        b.state_ref.conc_mol['S']*b.state_ref.conc_mol['E'])
            elif r == 'R2':
                return (b.reaction_rate[r] ==
                        b.reaction_coef[r] *
                        b.state_ref.conc_mol['C'])
            elif r == 'R3':
                return (b.reaction_rate[r] ==
                        b.reaction_coef[r] *
                        b.state_ref.conc_mol['C'])

        self.reaction_rate_eqn = Constraint(self._params.rate_reaction_idx,
                                            rule=reaction_rate_rule)

        def arrhenius_rule(b, r):
            return (b.reaction_coef[r] == b._params.k_rxn[r] *
                    exp(-b._params.act_energy[r]/b._params.gas_const /
                        b.state_ref.temperature))

        self.arrhenius_eqn = Constraint(self._params.rate_reaction_idx,
                                        rule=arrhenius_rule)

    def get_reaction_rate_basis(b):
        return MaterialFlowBasis.molar


@pytest.fixture
def model():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.sb = m.fs.pp.state_block_class(default={'parameters': m.fs.pp})

    for i in m.fs.sb.flow_mol_phase_comp:
        assert not m.fs.sb.flow_mol_phase_comp[i].fixed
        assert m.fs.sb.flow_mol_phase_comp[i].value == 2
    assert not m.fs.sb.pressure.fixed
    assert m.fs.sb.pressure.value == 1e5
    assert not m.fs.sb.temperature.fixed
    assert m.fs.sb.temperature.value == 300

    return m


@pytest.mark.unit
def test_fix_state_vars_basic(model):
    flags = fix_state_vars(model.fs.sb)

    for i in model.fs.sb.flow_mol_phase_comp:
        assert model.fs.sb.flow_mol_phase_comp[i].fixed
        assert model.fs.sb.flow_mol_phase_comp[i].value == 2
    assert model.fs.sb.pressure.fixed
    assert model.fs.sb.pressure.value == 1e5
    assert model.fs.sb.temperature.fixed
    assert model.fs.sb.temperature.value == 300

    assert not flags[None, "component_flow_phase", ("p1", "c1")]
    assert not flags[None, "component_flow_phase", ("p1", "c2")]
    assert not flags[None, "component_flow_phase", ("p2", "c1")]
    assert not flags[None, "component_flow_phase", ("p2", "c2")]
    assert not flags[None, "pressure", None]
    assert not flags[None, "temperature", None]


@pytest.mark.unit
def test_fix_state_vars_None_value(model):
    model.fs.sb.pressure.value = None

    with pytest.raises(ConfigurationError):
        fix_state_vars(model.fs.sb)


@pytest.mark.unit
def test_fix_state_vars_guesses(model):
    # Note that flow_mol_phase_comp is labled as component_flow
    # in define_state_vars
    state_args = {"component_flow_phase": {("p1", "c1"): 1,
                                           ("p1", "c2"): 2,
                                           ("p2", "c1"): 3,
                                           ("p2", "c2"): 4},
                  "pressure": 2e5,
                  "temperature": 500}

    flags = fix_state_vars(model.fs.sb, state_args)

    assert model.fs.sb.flow_mol_phase_comp[("p1", "c1")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c1")].value == 1
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c2")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c2")].value == 2
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c1")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c1")].value == 3
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c2")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c2")].value == 4

    assert model.fs.sb.pressure.fixed
    assert model.fs.sb.pressure.value == 2e5
    assert model.fs.sb.temperature.fixed
    assert model.fs.sb.temperature.value == 500

    assert not flags[None, "component_flow_phase", ("p1", "c1")]
    assert not flags[None, "component_flow_phase", ("p1", "c2")]
    assert not flags[None, "component_flow_phase", ("p2", "c1")]
    assert not flags[None, "component_flow_phase", ("p2", "c2")]
    assert not flags[None, "pressure", None]
    assert not flags[None, "temperature", None]


@pytest.mark.unit
def test_fix_state_vars_partial_guesses(model):
    # Note that flow_mol_phase_comp is labled as compoennt_flow
    # in define_state_vars
    state_args = {"component_flow_phase": {("p1", "c1"): 1,
                                           ("p1", "c2"): 2,
                                           ("p2", "c1"): 3,
                                           ("p2", "c2"): 4}}

    flags = fix_state_vars(model.fs.sb, state_args)

    assert model.fs.sb.flow_mol_phase_comp[("p1", "c1")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c1")].value == 1
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c2")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c2")].value == 2
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c1")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c1")].value == 3
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c2")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c2")].value == 4

    assert model.fs.sb.pressure.fixed
    assert model.fs.sb.pressure.value == 1e5
    assert model.fs.sb.temperature.fixed
    assert model.fs.sb.temperature.value == 300

    assert not flags[None, "component_flow_phase", ("p1", "c1")]
    assert not flags[None, "component_flow_phase", ("p1", "c2")]
    assert not flags[None, "component_flow_phase", ("p2", "c1")]
    assert not flags[None, "component_flow_phase", ("p2", "c2")]
    assert not flags[None, "pressure", None]
    assert not flags[None, "temperature", None]


@pytest.mark.unit
def test_fix_state_vars_guesses_mismatch_index(model):
    # Note that flow_mol_phase_comp is labled as compoennt_flow
    # in define_state_vars
    state_args = {"component_flow_phase": {("p1", "c1"): 1,
                                           ("p1", "c2"): 2,
                                           ("p2", "c1"): 3},
                  "pressure": 2e5,
                  "temperature": 500}

    with pytest.raises(ConfigurationError):
        fix_state_vars(model.fs.sb, state_args)


@pytest.mark.unit
def test_fix_state_vars_fixed_no_guesses(model):
    # Note that flow_mol_phase_comp is labled as compoennt_flow
    # in define_state_vars
    model.fs.sb.flow_mol_phase_comp["p1", "c1"].fix(10)
    model.fs.sb.pressure.fix(1.5e5)

    flags = fix_state_vars(model.fs.sb)

    assert model.fs.sb.flow_mol_phase_comp[("p1", "c1")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c1")].value == 10
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c2")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c2")].value == 2
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c1")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c1")].value == 2
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c2")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c2")].value == 2

    assert model.fs.sb.pressure.fixed
    assert model.fs.sb.pressure.value == 1.5e5
    assert model.fs.sb.temperature.fixed
    assert model.fs.sb.temperature.value == 300

    # Pressure and component_flow_phase[p1, c1] should be True
    assert flags[None, "component_flow_phase", ("p1", "c1")]
    assert not flags[None, "component_flow_phase", ("p1", "c2")]
    assert not flags[None, "component_flow_phase", ("p2", "c1")]
    assert not flags[None, "component_flow_phase", ("p2", "c2")]
    assert flags[None, "pressure", None]
    assert not flags[None, "temperature", None]


@pytest.mark.unit
def test_fix_state_vars_fixed_guesses(model):
    # Note that flow_mol_phase_comp is labled as compoennt_flow
    # in define_state_vars
    model.fs.sb.flow_mol_phase_comp["p1", "c1"].fix(10)
    model.fs.sb.pressure.fix(1.5e5)

    state_args = {"component_flow_phase": {("p1", "c1"): 1,
                                           ("p1", "c2"): 2,
                                           ("p2", "c1"): 3,
                                           ("p2", "c2"): 4},
                  "pressure": 2e5,
                  "temperature": 500}

    flags = fix_state_vars(model.fs.sb, state_args)

    assert model.fs.sb.flow_mol_phase_comp[("p1", "c1")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c1")].value == 10
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c2")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c2")].value == 2
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c1")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c1")].value == 3
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c2")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c2")].value == 4

    assert model.fs.sb.pressure.fixed
    assert model.fs.sb.pressure.value == 1.5e5
    assert model.fs.sb.temperature.fixed
    assert model.fs.sb.temperature.value == 500

    # Pressure and component_flow_phase[p1, c1] should be True
    assert flags[None, "component_flow_phase", ("p1", "c1")]
    assert not flags[None, "component_flow_phase", ("p1", "c2")]
    assert not flags[None, "component_flow_phase", ("p2", "c1")]
    assert not flags[None, "component_flow_phase", ("p2", "c2")]
    assert flags[None, "pressure", None]
    assert not flags[None, "temperature", None]


@pytest.mark.unit
def test_revert_state_vars_basic(model):
    flags = fix_state_vars(model.fs.sb)

    revert_state_vars(model.fs.sb, flags)

    for i in model.fs.sb.flow_mol_phase_comp:
        assert not model.fs.sb.flow_mol_phase_comp[i].fixed
        assert model.fs.sb.flow_mol_phase_comp[i].value == 2
    assert not model.fs.sb.pressure.fixed
    assert model.fs.sb.pressure.value == 1e5
    assert not model.fs.sb.temperature.fixed
    assert model.fs.sb.temperature.value == 300


@pytest.mark.unit
def test_revert_state_vars_guesses(model):
    # Note that flow_mol_phase_comp is labled as compoennt_flow
    # in define_state_vars
    state_args = {"component_flow_phase": {("p1", "c1"): 1,
                                           ("p1", "c2"): 2,
                                           ("p2", "c1"): 3,
                                           ("p2", "c2"): 4},
                  "pressure": 2e5,
                  "temperature": 500}

    flags = fix_state_vars(model.fs.sb, state_args)

    revert_state_vars(model.fs.sb, flags)

    assert not model.fs.sb.flow_mol_phase_comp[("p1", "c1")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c1")].value == 1
    assert not model.fs.sb.flow_mol_phase_comp[("p1", "c2")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c2")].value == 2
    assert not model.fs.sb.flow_mol_phase_comp[("p2", "c1")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c1")].value == 3
    assert not model.fs.sb.flow_mol_phase_comp[("p2", "c2")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c2")].value == 4

    assert not model.fs.sb.pressure.fixed
    assert model.fs.sb.pressure.value == 2e5
    assert not model.fs.sb.temperature.fixed
    assert model.fs.sb.temperature.value == 500


@pytest.mark.unit
def test_revert_state_vars_fixed_no_guesses(model):
    # Note that flow_mol_phase_comp is labled as compoennt_flow
    # in define_state_vars
    model.fs.sb.flow_mol_phase_comp["p1", "c1"].fix(10)
    model.fs.sb.pressure.fix(1.5e5)

    flags = fix_state_vars(model.fs.sb)
    revert_state_vars(model.fs.sb, flags)

    # Pressure and componet_flow[p1, c1] should still be fixed
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c1")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c1")].value == 10
    assert not model.fs.sb.flow_mol_phase_comp[("p1", "c2")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c2")].value == 2
    assert not model.fs.sb.flow_mol_phase_comp[("p2", "c1")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c1")].value == 2
    assert not model.fs.sb.flow_mol_phase_comp[("p2", "c2")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c2")].value == 2

    assert model.fs.sb.pressure.fixed
    assert model.fs.sb.pressure.value == 1.5e5
    assert not model.fs.sb.temperature.fixed
    assert model.fs.sb.temperature.value == 300


@pytest.mark.unit
def test_revert_state_vars_fixed_guesses(model):
    # Note that flow_mol_phase_comp is labled as compoennt_flow
    # in define_state_vars
    model.fs.sb.flow_mol_phase_comp["p1", "c1"].fix(10)
    model.fs.sb.pressure.fix(1.5e5)

    state_args = {"component_flow_phase": {("p1", "c1"): 1,
                                           ("p1", "c2"): 2,
                                           ("p2", "c1"): 3,
                                           ("p2", "c2"): 4},
                  "pressure": 2e5,
                  "temperature": 500}

    flags = fix_state_vars(model.fs.sb, state_args)
    revert_state_vars(model.fs.sb, flags)

    # Pressure and componet_flow[p1, c1] should still be fixed
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c1")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c1")].value == 10
    assert not model.fs.sb.flow_mol_phase_comp[("p1", "c2")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c2")].value == 2
    assert not model.fs.sb.flow_mol_phase_comp[("p2", "c1")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c1")].value == 3
    assert not model.fs.sb.flow_mol_phase_comp[("p2", "c2")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c2")].value == 4

    assert model.fs.sb.pressure.fixed
    assert model.fs.sb.pressure.value == 1.5e5
    assert not model.fs.sb.temperature.fixed
    assert model.fs.sb.temperature.value == 500


@pytest.mark.unit
def test_revert_state_vars_flag_mismatch(model):
    flags = {(None, 'component_flow_phase', ('p1', 'c1')): False,
             (None, 'component_flow_phase', ('p1', 'c2')): False,
             (None, 'component_flow_phase', ('p2', 'c1')): False,
             (None, 'pressure', None): False}

    with pytest.raises(ConfigurationError):
        revert_state_vars(model.fs.sb, flags)


@pytest.mark.unit
def test_propagate_state():
    m = ConcreteModel()

    def block_rule(b):
        b.s = Set(initialize=[1, 2])
        b.v1 = Var()
        b.v2 = Var(b.s)

        b.p = Port()
        b.p.add(b.v1, "V1")
        b.p.add(b.v2, "V2")
        return

    m.b1 = Block(rule=block_rule)
    m.b2 = Block(rule=block_rule)

    m.s1 = Arc(source=m.b1.p, destination=m.b2.p)

    # Set values on first block
    m.b1.v1.value = 10
    m.b1.v2[1].value = 20
    m.b1.v2[2].value = 30

    # Make sure vars in block 2 haven't been changed accidentally
    assert m.b2.v1.value is None
    assert m.b2.v2[1].value is None
    assert m.b2.v2[2].value is None

    propagate_state(m.s1)

    # Check that values were propagated correctly
    assert m.b2.v1.value == m.b1.v1.value
    assert m.b2.v2[1].value == m.b1.v2[1].value
    assert m.b2.v2[2].value == m.b1.v2[2].value

    assert m.b1.v1.fixed is False
    assert m.b1.v2[1].fixed is False
    assert m.b1.v2[2].fixed is False
    assert m.b2.v1.fixed is False
    assert m.b2.v2[1].fixed is False
    assert m.b2.v2[2].fixed is False


@pytest.mark.unit
def test_propagate_state_reverse():
    m = ConcreteModel()

    def block_rule(b):
        b.s = Set(initialize=[1, 2])
        b.v1 = Var()
        b.v2 = Var(b.s)

        b.p = Port()
        b.p.add(b.v1, "V1")
        b.p.add(b.v2, "V2")
        return

    m.b1 = Block(rule=block_rule)
    m.b2 = Block(rule=block_rule)

    m.s1 = Arc(source=m.b1.p, destination=m.b2.p)

    # Test reverse propogation - set values on second block
    m.b2.v1.value = 100
    m.b2.v2[1].value = 200
    m.b2.v2[2].value = 300

    # Make sure vars in block 1 haven't been changed accidentally
    assert m.b1.v1.value is None
    assert m.b1.v2[1].value is None
    assert m.b1.v2[2].value is None

    propagate_state(m.s1, direction="backward")

    # Check that values were propagated correctly
    assert m.b2.v1.value == m.b1.v1.value
    assert m.b2.v2[1].value == m.b1.v2[1].value
    assert m.b2.v2[2].value == m.b1.v2[2].value

    assert m.b1.v1.fixed is False
    assert m.b1.v2[1].fixed is False
    assert m.b1.v2[2].fixed is False
    assert m.b2.v1.fixed is False
    assert m.b2.v2[1].fixed is False
    assert m.b2.v2[2].fixed is False


@pytest.mark.unit
def test_propagate_state_fixed():
    m = ConcreteModel()

    def block_rule(b):
        b.s = Set(initialize=[1, 2])
        b.v1 = Var()
        b.v2 = Var(b.s)

        b.p = Port()
        b.p.add(b.v1, "V1")
        b.p.add(b.v2, "V2")
        return

    m.b1 = Block(rule=block_rule)
    m.b2 = Block(rule=block_rule)

    m.s1 = Arc(source=m.b1.p, destination=m.b2.p)

    # Set values on first block
    m.b1.v1.value = 10
    m.b1.v2[1].value = 20
    m.b1.v2[2].value = 30

    # Make sure vars in block 2 haven't been changed accidentally
    assert m.b2.v1.value is None
    assert m.b2.v2[1].value is None
    assert m.b2.v2[2].value is None

    # Fix v1 in block 2
    m.b2.v1.fix(500)

    propagate_state(m.s1)

    # Check that values were propagated correctly
    assert m.b2.v1.value == 500
    assert m.b2.v2[1].value == m.b1.v2[1].value
    assert m.b2.v2[2].value == m.b1.v2[2].value

    assert m.b1.v1.fixed is False
    assert m.b1.v2[1].fixed is False
    assert m.b1.v2[2].fixed is False
    assert m.b2.v1.fixed is True
    assert m.b2.v2[1].fixed is False
    assert m.b2.v2[2].fixed is False


@pytest.mark.unit
def test_propagate_state_Expression():
    m = ConcreteModel()

    def block_rule(b):
        b.s = Set(initialize=[1, 2])
        b.v1 = Var()
        b.v2 = Var(b.s)

        b.e = Expression(expr=b.v1)

        b.p = Port()
        b.p.add(b.e, "E")
        b.p.add(b.v2, "V2")
        return

    m.b1 = Block(rule=block_rule)
    m.b2 = Block(rule=block_rule)

    m.s1 = Arc(source=m.b1.p, destination=m.b2.p)

    with pytest.raises(TypeError):
        propagate_state(m.s1)


@pytest.mark.unit
def test_propagate_state_invalid_stream():
    m = ConcreteModel()

    with pytest.raises(TypeError):
        propagate_state(m)


@pytest.mark.unit
def test_propagate_state_invalid_direction():
    m = ConcreteModel()

    def block_rule(b):
        b.s = Set(initialize=[1, 2])
        b.v1 = Var()
        b.v2 = Var(b.s)

        b.p = Port()
        b.p.add(b.v1, "V1")
        b.p.add(b.v2, "V2")
        return

    m.b1 = Block(rule=block_rule)
    m.b2 = Block(rule=block_rule)

    m.s1 = Arc(source=m.b1.p, destination=m.b2.p)

    with pytest.raises(ValueError):
        propagate_state(m.s1, direction="foo")


@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.unit
def test_solve_indexed_block_list():
    # Create an indexed block and try to solve it
    m = ConcreteModel()
    m.s = Set(initialize=[1, 2, 3])

    def block_rule(b, x):
        b.v = Var(initialize=1.0)
        b.c = Constraint(expr=b.v == 2.0)
    m.b = Block(m.s, rule=block_rule)

    solve_indexed_blocks(solver=solver, blocks=[m.b])

    for i in m.s:
        assert value(m.b[i].v == 2.0)


@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.unit
def test_solve_indexed_block_IndexedBlock():
    # Create an indexed block and try to solve it
    m = ConcreteModel()
    m.s = Set(initialize=[1, 2, 3])

    def block_rule(b, x):
        b.v = Var(initialize=1.0)
        b.c = Constraint(expr=b.v == 2.0)
    m.b = Block(m.s, rule=block_rule)

    solve_indexed_blocks(solver=solver, blocks=m.b)

    for i in m.s:
        assert value(m.b[i].v == 2.0)


@pytest.mark.unit
def test_solve_indexed_block_error():
    # Try solve_indexed_block on non-block object
    with pytest.raises(TypeError):
        solve_indexed_blocks(solver=None, blocks=[1, 2, 3])


@pytest.mark.integration
@pytest.mark.skipif(solver is None, reason="Solver not available")
def test_initialize_by_time_element():
    horizon = 6
    time_set = [0, horizon]
    ntfe = 60  # For a finite element every six seconds
    ntcp = 2
    m = ConcreteModel(name='CSTR model for testing')
    m.fs = FlowsheetBlock(default={'dynamic': True,
                                   'time_set': time_set})

    m.fs.properties = AqueousEnzymeParameterBlock()
    m.fs.reactions = EnzymeReactionParameterBlock(
            default={'property_package': m.fs.properties})
    m.fs.cstr = CSTR(default={
        "property_package": m.fs.properties,
        "reaction_package": m.fs.reactions,
        "material_balance_type": MaterialBalanceType.componentTotal,
        "energy_balance_type": EnergyBalanceType.enthalpyTotal,
        "momentum_balance_type": MomentumBalanceType.none,
        "has_heat_of_reaction": True})

    # Time discretization
    disc = TransformationFactory('dae.collocation')
    disc.apply_to(
        m, wrt=m.fs.time, nfe=ntfe, ncp=ntcp, scheme='LAGRANGE-RADAU')

    # Fix geometry variables
    m.fs.cstr.volume[0].fix(1.0)

    # Fix initial conditions:
    for p, j in m.fs.properties.phase_list*m.fs.properties.component_list:
        if j == 'Solvent':
            continue
        m.fs.cstr.control_volume.material_holdup[0, p, j].fix(0)

    # Fix inlet conditions
    # This is a huge hack because I didn't know that the proper way to
    # have multiple inlets to a CSTR was to use a mixer.
    # I 'combine' both my inlet streams before sending them to the CSTR.
    for t, j in m.fs.time*m.fs.properties.component_list:
        if t <= 2:
            if j == 'E':
                m.fs.cstr.inlet.conc_mol[t, j].fix(11.91*0.1/2.2)
            elif j == 'S':
                m.fs.cstr.inlet.conc_mol[t, j].fix(12.92*2.1/2.2)
            elif j != 'Solvent':
                m.fs.cstr.inlet.conc_mol[t, j].fix(0)
        elif t <= 4:
            if j == 'E':
                m.fs.cstr.inlet.conc_mol[t, j].fix(5.95*0.1/2.2)
            elif j == 'S':
                m.fs.cstr.inlet.conc_mol[t, j].fix(12.92*2.1/2.2)
            elif j != 'Solvent':
                m.fs.cstr.inlet.conc_mol[t, j].fix(0)
        else:
            if j == 'E':
                m.fs.cstr.inlet.conc_mol[t, j].fix(8.95*0.1/2.2)
            elif j == 'S':
                m.fs.cstr.inlet.conc_mol[t, j].fix(16.75*2.1/2.2)
            elif j != 'Solvent':
                m.fs.cstr.inlet.conc_mol[t, j].fix(0)

    m.fs.cstr.inlet.conc_mol[:, 'Solvent'].fix(1.)

    m.fs.cstr.inlet.flow_vol.fix(2.2)
    m.fs.cstr.inlet.temperature.fix(300)

    # Fix outlet conditions
    m.fs.cstr.outlet.flow_vol.fix(2.2)
    m.fs.cstr.outlet.temperature[m.fs.time.first()].fix(300)

    assert degrees_of_freedom(m) == 0

    initialize_by_time_element(m.fs, m.fs.time, solver=solver)

    assert degrees_of_freedom(m) == 0

    # Assert that the result looks how we expect
    assert m.fs.cstr.outlet.conc_mol[0, 'S'].value == 0
    assert abs(m.fs.cstr.outlet.conc_mol[2, 'S'].value - 11.389) < 1e-2
    assert abs(m.fs.cstr.outlet.conc_mol[4, 'P'].value - 0.2191) < 1e-3
    assert abs(m.fs.cstr.outlet.conc_mol[6, 'E'].value - 0.0327) < 1e-3
    assert abs(m.fs.cstr.outlet.temperature[6].value - 289.7) < 1

    # Assert that model is still fixed and deactivated as expected
    assert (
        m.fs.cstr.control_volume.material_holdup[
            m.fs.time.first(), 'aq', 'S'].fixed)

    for t in m.fs.time:
        if t != m.fs.time.first():
            assert (not m.fs.cstr.control_volume.material_holdup[
                t, 'aq', 'S'].fixed)

            assert not m.fs.cstr.outlet.temperature[t].fixed
        assert (m.fs.cstr.control_volume.material_holdup_calculation[
            t, 'aq', 'C'].active)

        assert m.fs.cstr.control_volume.properties_out[t].active
        assert not m.fs.cstr.outlet.conc_mol[t, 'S'].fixed
        assert m.fs.cstr.inlet.conc_mol[t, 'S'].fixed

    # Assert that constraints are feasible after initialization
    for con in m.fs.component_data_objects(Constraint, active=True):
        assert value(con.body) - value(con.upper) < 1e-5
        assert value(con.lower) - value(con.body) < 1e-5

    results = solver.solve(m.fs)
    assert results.solver.termination_condition == TerminationCondition.optimal
